import logging
import os
import select
import time
from threading import Event as TEvent


from hhd.controller import Multiplexer, DEBUG_MODE
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev, enumerate_evs
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.controller.virtual.uinput import UInputDevice
from hhd.controller.physical.hidraw import enumerate_unique
from hhd.plugins import Config, Context, Emitter, get_gyro_state, get_outputs
from .serial import SerialDevice, get_serial
from .hid_v1 import OxpHidraw
from .hid_v2 import OxpHidrawV2
from .const import BTN_MAPPINGS, DEFAULT_MAPPINGS, BTN_MAPPINGS_NONTURBO

FIND_DELAY = 0.1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3
TURBO_DELAY = 5
TURBO_CONTROLLER_CHECK = 2

logger = logging.getLogger(__name__)


GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

KBD_VID = 0x0001
KBD_PID = 0x0001

X1_MINI_VID = 0x1A86
X1_MINI_PID = 0xFE00
X1_MINI_PAGE = 0xFF00
X1_MINI_USAGE = 0x0001

XFLY_VID = 0x1A2C
XFLY_PID = 0xB001
XFLY_PAGE = 0xFF01
XFLY_USAGE = 0x0001

BACK_BUTTON_DELAY = 0.1

RGB_MODES_FULL = {
    "disabled": [],
    "oxp": ["oxp", "oxp-secondary"],
    "solid": ["color"],
    "duality": ["dual"],
}
RGB_MODES_STICKS = {
    "disabled": [],
    "oxp": ["oxp"],
    "solid": ["color"],
}


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    turbo: bool = False,
):
    first = True
    init = time.perf_counter()
    repeated_fail = False
    switch_to_turbo = None
    first_disabled = True

    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            if first_disabled:
                UInputDevice.close_volume_cached()
                unhide_all()
            first_disabled = False
            continue
        else:
            first_disabled = True

        try:
            found_device = bool(enumerate_evs(vid=GAMEPAD_VID, pid=GAMEPAD_PID))
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            time.sleep(LONGER_ERROR_DELAY)
            found_device = True

        try:
            protocol = dconf.get("protocol", None)
            # Serial device is always present
            # Hid devices might not be, wait a bit for them
            match protocol:
                case "hid_v1":
                    found_vendor = bool(
                        enumerate_unique(
                            vid=X1_MINI_VID,
                            pid=X1_MINI_PID,
                            usage_page=X1_MINI_PAGE,
                            usage=X1_MINI_USAGE,
                        )
                    )
                case "hid_v2":
                    found_vendor = bool(
                        enumerate_unique(
                            vid=XFLY_VID,
                            pid=XFLY_PID,
                            usage_page=XFLY_PAGE,
                            usage=XFLY_USAGE,
                        )
                    )
                case "hid_v1_g1":
                    found_vendor = bool(
                        enumerate_unique(
                            vid=XFLY_VID,
                            pid=XFLY_PID,
                            usage_page=XFLY_PAGE,
                            usage=XFLY_USAGE,
                        )
                    )
                case "mixed":
                    found_vendor = bool(
                        enumerate_unique(
                            vid=XFLY_VID,
                            pid=XFLY_PID,
                            usage_page=XFLY_PAGE,
                            usage=XFLY_USAGE,
                        )
                    ) and bool(get_serial()[0])
                case "serial":
                    found_vendor = bool(get_serial()[0])
                case _:
                    found_vendor = True
        except Exception:
            logger.warning("Failed finding vendor device, skipping check.")
            found_vendor = True

        turbo_start = False
        if not found_device or not found_vendor:
            curr = time.perf_counter()
            if first:
                logger.info("Controller not found. Waiting...")
                switch_to_turbo = curr + TURBO_DELAY
            time.sleep(FIND_DELAY)
            first = False
            if found_vendor and turbo and switch_to_turbo and curr > switch_to_turbo:
                logger.info("Switching to turbo only button mode")
                updated.clear()
                turbo_start = True
                first = False
            else:
                continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            if turbo_start:
                turbo_loop(conf.copy(), should_exit, updated, dconf, emit)
            else:
                controller_loop(conf.copy(), should_exit, updated, dconf, emit, turbo)
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            first = True
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)

    # Close the volume keyboard cache
    UInputDevice.close_volume_cached()
    unhide_all()


def find_vendor(prepare, turbo, protocol: str | None):
    d_ser = SerialDevice(turbo=turbo, required=True)
    d_hidraw = OxpHidraw(
        vid=[X1_MINI_VID],
        pid=[X1_MINI_PID],
        usage_page=[X1_MINI_PAGE],
        usage=[X1_MINI_USAGE],
        turbo=turbo,
        required=True,
    )
    d_hidraw_v2 = OxpHidrawV2(
        vid=[XFLY_VID],
        pid=[XFLY_PID],
        usage_page=[XFLY_PAGE],
        usage=[XFLY_USAGE],
        turbo=turbo,
        required=True,
    )
    d_hidraw_g1 = OxpHidraw(
        vid=[XFLY_VID],
        pid=[XFLY_PID],
        usage_page=[XFLY_PAGE],
        usage=[XFLY_USAGE],
        turbo=turbo,
        required=True,
        g1=True,
    )

    if protocol in ["serial", "mixed"]:
        try:
            prepare(d_ser)
            # OneXFly uses serial only for the buttons and hidraw for RGB
            # Initialize V2 selectcively on that one
            try:
                if d_ser.buttons_only:
                    if protocol == "serial":
                        logger.warning(
                            f"Device has protocol 'serial', but 'mixed' was detected."
                        )
                    prepare(d_hidraw_v2)
                return [d_ser, d_hidraw_v2]
            except Exception as e:
                logger.info(
                    f"Could not find V2 hidraw vendor device, RGB will not work, error:\n{e}"
                )
                return [d_ser]
        except Exception as e:
            pass

    if protocol == "hid_v1":
        try:
            prepare(d_hidraw)
            logger.info("Found OXP V1 hidraw vendor device.")
            return [d_hidraw]
        except Exception as e:
            pass

    if protocol == "hid_v1_g1":
        try:
            prepare(d_hidraw_g1)
            logger.info("Found OXP V1 hidraw vendor device.")
            return [d_hidraw_g1]
        except Exception as e:
            pass

    if protocol == "hid_v2":
        try:
            prepare(d_hidraw_v2)
            logger.info("Found OXP V2 hidraw vendor device.")
            return [d_hidraw_v2]
        except Exception as e:
            pass

    logger.error("No vendor device found, RGB and back buttons will not work.")
    return []


def turbo_loop(
    conf: Config,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    emit: Emitter,
):
    debug = DEBUG_MODE

    # Output
    if dconf.get("rgb_secondary", False):
        rgb_modes = RGB_MODES_FULL
    elif dconf.get("rgb", True):
        rgb_modes = RGB_MODES_STICKS
    else:
        rgb_modes = None

    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        conf["imu"].to(bool),
        emit=emit,
        rgb_modes=rgb_modes,  # type: ignore
        controller_disabled=True,
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        grab=True,
        btn_map=BTN_MAPPINGS,
    )

    share_reboots = False
    last_controller_check = 0
    keyboard_is = "keyboard"
    qam_hhd = False
    qam_no_release = False
    if conf.get("turbo_reboots", False):
        share_reboots = True
    
    if not dconf.get("g1", False):
        match conf.get("extra_buttons", "separate"):
            case "separate":
                keyboard_is = "steam_qam"
                qam_hhd = True
            case "combo":
                keyboard_is = "qam"
                qam_hhd = False
            case "combo_hhd":
                keyboard_is = "qam"
                qam_hhd = True

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        params=d_params,
        share_reboots=share_reboots,
        keyboard_is=keyboard_is,
        swap_guide="start_is_keyboard" if conf.get("swap_face", False) else None,
        qam_hhd=qam_hhd,
        qam_no_release=qam_no_release,
        keyboard_no_release=not conf.get("swap_face", False),
    )

    if dconf.get("x1", False) and conf.get("volume_reverse", False):
        logger.info("Reversing volume buttons.")
        btn_map = {
            "key_volumedown": EC("KEY_VOLUMEUP"),
            "key_volumeup": EC("KEY_VOLUMEDOWN"),
        }
    else:
        btn_map = {
            "key_volumeup": EC("KEY_VOLUMEUP"),
            "key_volumedown": EC("KEY_VOLUMEDOWN"),
        }

    d_volume_btn = UInputDevice(
        name="Handheld Daemon Volume Keyboard",
        phys="phys-hhd-vbtn",
        capabilities={EC("EV_KEY"): [EC("KEY_VOLUMEUP"), EC("KEY_VOLUMEDOWN")]},
        btn_map=btn_map,  # type: ignore
        pid=KBD_PID,
        vid=KBD_VID,
        output_timestamps=True,
        volume_keyboard=True,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 25

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        prepare(d_volume_btn)
        d_vend = find_vendor(prepare, True, dconf.get("protocol", None))
        d_vend_id = [id(d) for d in d_vend]

        for d in d_producers:
            prepare(d)
        prepare(d_kbd_1)

        logger.info(
            "Turbo only mode started, the turbo button of the device will still work."
        )
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()

            if start - last_controller_check > TURBO_CONTROLLER_CHECK:
                last_controller_check = start
                try:
                    found_device = bool(enumerate_evs(vid=GAMEPAD_VID, pid=GAMEPAD_PID))
                except Exception:
                    logger.warning("Failed finding device, skipping check.")
                    found_device = True
                if found_device:
                    logger.info("Controller found, switching to controller mode.")
                    break

            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                d_id = id(d)
                if d_id in to_run or d_id in d_vend_id:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_volume_btn.consume(evs)

            for d in d_vend:
                d.consume(evs)
            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e


def controller_loop(
    conf: Config,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    emit: Emitter,
    turbo: bool = False,
):
    debug = DEBUG_MODE

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        conf["imu"].to(bool),
        emit=emit,
        rgb_modes=(
            RGB_MODES_FULL if dconf.get("rgb_secondary", False) else RGB_MODES_STICKS  # type: ignore
        ),
    )
    motion = d_params.get("uses_motion", True)

    # Imu
    d_imu = CombinedImu(
        conf["imu_hz"].to(int),
        get_gyro_state(conf["imu_axis"], dconf.get("mapping", DEFAULT_MAPPINGS)),
    )
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID],
        pid=[GAMEPAD_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )

    if turbo:
        # Switch buttons if turbo is enabled.
        # This only affects AOKZOE and OneXPlayer devices with
        # that button that have the nonturbo mapping as default
        mappings = BTN_MAPPINGS
    else:
        mappings = BTN_MAPPINGS_NONTURBO

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        grab=True,
        btn_map=mappings,
    )
    # Touchpad keyboard
    d_kbd_2 = GenericGamepadEvdev(
        vid=[0x6080],
        pid=[0x8060],
        required=True,
        grab=False,
        btn_map=BTN_MAPPINGS,
        capabilities={EC("EV_KEY"): [EC("KEY_D")]},
        requires_start=True,
    )

    share_reboots = False
    keyboard_is = "keyboard"
    qam_hhd = False
    qam_no_release = False
    if turbo:
        if conf.get("turbo_reboots", False):
            share_reboots = True
        match conf.get("extra_buttons", "separate"):
            case "separate":
                keyboard_is = "steam_qam"
                qam_hhd = True
            case "combo":
                keyboard_is = "qam"
                qam_hhd = False
            case "combo_hhd":
                keyboard_is = "qam"
                qam_hhd = True
    else:
        qam_no_release = True

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        params=d_params,
        share_reboots=share_reboots,
        keyboard_is=keyboard_is,
        swap_guide="start_is_keyboard" if conf.get("swap_face", False) else None,
        qam_hhd=qam_hhd,
        qam_no_release=qam_no_release,
        keyboard_no_release=not conf.get("swap_face", False),
    )

    if dconf.get("x1", False) and conf.get("volume_reverse", False):
        logger.info("Reversing volume buttons.")
        btn_map = {
            "key_volumedown": EC("KEY_VOLUMEUP"),
            "key_volumeup": EC("KEY_VOLUMEDOWN"),
        }
    else:
        btn_map = {
            "key_volumeup": EC("KEY_VOLUMEUP"),
            "key_volumedown": EC("KEY_VOLUMEDOWN"),
        }

    d_volume_btn = UInputDevice(
        name="Handheld Daemon Volume Keyboard",
        phys="phys-hhd-vbtn",
        capabilities={EC("EV_KEY"): [EC("KEY_VOLUMEUP"), EC("KEY_VOLUMEDOWN")]},
        btn_map=btn_map,  # type: ignore
        pid=KBD_PID,
        vid=KBD_VID,
        output_timestamps=True,
        volume_keyboard=True,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 125

    if motion:
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        d_vend = find_vendor(prepare, turbo, dconf.get("protocol", None))
        d_vend_id = [id(d) for d in d_vend]
        if dconf.get("g1", False):
            prepare(d_kbd_2)
        prepare(d_xinput)
        if motion:
            start_imu = True
            if dconf.get("hrtimer", False):
                start_imu = d_timer.open()
            if start_imu:
                prepare(d_imu)
        prepare(d_volume_btn)
        prepare(d_kbd_1)

        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                d_id = id(d)
                if d_id in to_run or d_id in d_vend_id:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_volume_btn.consume(evs)
                d_xinput.consume(evs)

            for d in d_vend:
                d.consume(evs)
            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        # d_vend.close(not updated.is_set())
        try:
            d_timer.close()
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



import logging
import re
import select
import time
from threading import Event as TEvent
from typing import Sequence

import evdev

from hhd.controller import DEBUG_MODE, Event, Multiplexer, can_read
from hhd.controller.base import Event
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.plugins import Config, Context, Emitter, get_gyro_state, get_outputs

from .const import (
    GPD_TOUCHPAD_AXIS_MAP,
    GPD_TOUCHPAD_BUTTON_MAP,
    GPD_WIN_DEFAULT_MAPPINGS,
)

ERROR_DELAY = 0.3
SELECT_TIMEOUT = 1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3

logger = logging.getLogger(__name__)

# Old devices were 2f24:0135
# New 2025 Win mini uses 045e:002d
GPD_WIN_VIDS = [0x2F24, 0x045E]
GPD_WIN_PIDS = [0x0135, 0x002D]
GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

# Win Max 2
TOUCHPAD_VID = 0x093A
TOUCHPAD_PID = 0x0255
# Win Minis
TOUCHPAD_VID_2 = 0x0911
TOUCHPAD_PID_2 = 0x5288

BACK_BUTTON_DELAY = 0.025

# /dev/input/event17 Microsoft X-Box 360 pad usb-0000:73:00.3-4.1/input0
# bus: 0003, vendor 045e, product 028e, version 0101

# back buttons
# /dev/input/event15   Mouse for Windows usb-0000:73:00.3-4.2/input1
# bus: 0003, vendor 2f24, product 0135, version 0110

# physical keyboard
# /dev/input/event13   Mouse for Windows usb-0000:73:00.3-4.2/input0
# bus: 0003, vendor 2f24, product 0135, version 0110

# hidraw back buttons  {'path': b'/dev/hidraw6',
#    'vendor_id': 12068, 'product_id': 309, 'serial_number': '',
#    'release_number': 256, 'manufacturer_string': ' ',
#    'product_string': 'Mouse for Windows',
#    'usage_page': 1, 'usage': 6, 'interface_number': 1},

LEFT_BUTTONS = {
    EC("KEY_SYSRQ"),
    EC("KEY_F20"),
}

RIGHT_BUTTONS = {
    EC("KEY_PAUSE"),
    EC("KEY_F21"),
}


class BackbuttonsEvdev(GenericGamepadEvdev):
    def __init__(self, *args, **kwargs) -> None:
        self.left_pressed = False
        self.left_released = None
        self.right_pressed = False
        self.right_released = None
        super().__init__(*args, **kwargs)

    def produce(self, fds: Sequence[int]):
        if not self.dev:
            return []

        # GPD events execute micro sequences
        # Inbetween the sequences, there is a ~20ms gap in which the
        # button is not pressed. Therefore, record when the button was
        # pressed and if more than ~25ms has passed, consider it released.
        curr = time.perf_counter()
        out = []
        while self.fd in fds and can_read(self.dev):
            for e in self.dev.read():
                if e.type != EC("EV_KEY"):
                    continue

                pressed = e.value != 0

                if e.code in LEFT_BUTTONS:
                    if pressed:
                        if not self.left_pressed:
                            out.append(
                                {"type": "button", "code": "extra_l1", "value": True}
                            )
                        self.left_pressed = True
                        self.left_released = None
                    else:
                        self.left_released = curr
                if e.code in RIGHT_BUTTONS:
                    if pressed:
                        if not self.right_pressed:
                            out.append(
                                {
                                    "type": "button",
                                    "code": "extra_r1",
                                    "value": True,
                                }
                            )
                        self.right_pressed = True
                        self.right_released = None
                    else:
                        self.right_released = curr

        if self.left_released and curr - self.left_released > BACK_BUTTON_DELAY:
            out.append({"type": "button", "code": "extra_l1", "value": False})
            self.left_released = None
            self.left_pressed = False
        if self.right_released and curr - self.right_released > BACK_BUTTON_DELAY:
            out.append({"type": "button", "code": "extra_r1", "value": False})
            self.right_released = None
            self.right_pressed = False

        return out


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
):
    first = True
    first_disabled = True
    init = time.perf_counter()
    repeated_fail = False
    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            if first_disabled:
                unhide_all()
            first_disabled = False
            continue
        else:
            first_disabled = True

        found_gamepad = False
        try:
            for d in evdev.list_devices():
                dev = evdev.InputDevice(d)
                if dev.info.vendor == GAMEPAD_VID and dev.info.product == GAMEPAD_PID:
                    found_gamepad = True
                    break
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            found_gamepad = True

        if not found_gamepad:
            if first:
                logger.info("Controller in Mouse mode. Waiting...")
            time.sleep(ERROR_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            controller_loop(conf.copy(), should_exit, updated, dconf, emit)
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.exception(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            # Raise exception
            if DEBUG_MODE:
                import traceback

                logger.error(traceback.format_exc())
            time.sleep(sleep_time)

    # Unhide all devices before exiting
    unhide_all()


def controller_loop(
    conf: Config, should_exit: TEvent, updated: TEvent, dconf: dict, emit: Emitter
):
    debug = DEBUG_MODE
    has_touchpad = dconf.get("touchpad", False)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        conf["touchpad"] if has_touchpad else None,
        conf["imu"].to(bool),
        emit=emit,
    )
    motion = d_params.get("uses_motion", True)

    # Imu
    d_imu = CombinedImu(
        conf["imu_hz"].to(int),
        get_gyro_state(
            conf["imu_axis"], dconf.get("mapping", GPD_WIN_DEFAULT_MAPPINGS)
        ),
    )
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID],
        pid=[GAMEPAD_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )

    # "PNP0C50:00 0911:5288 Touchpad" on Win Max 2 2023
    # "PNP0C50:00 093A:0255 Touchpad" on Win Mini
    d_touch = GenericGamepadEvdev(
        vid=[TOUCHPAD_VID, TOUCHPAD_VID_2],
        pid=[TOUCHPAD_PID, TOUCHPAD_PID_2],
        name=[re.compile(".+Touchpad")],
        capabilities={EC("EV_KEY"): [EC("BTN_MOUSE")]},
        btn_map=GPD_TOUCHPAD_BUTTON_MAP,
        axis_map=GPD_TOUCHPAD_AXIS_MAP,
        aspect_ratio=1.333,
        required=False,
    )

    # Vendor
    d_kbd_1 = BackbuttonsEvdev(
        vid=GPD_WIN_VIDS,
        pid=GPD_WIN_PIDS,
        capabilities={EC("EV_KEY"): [EC("KEY_SYSRQ"), EC("KEY_PAUSE")]},
        required=True,
        grab=True,
        # btn_map={EC("KEY_SYSRQ"): "extra_l1", EC("KEY_PAUSE"): "extra_r1"},
    )

    match conf["l4r4"].to(str):
        case "l4":
            qam_button = "extra_l1"
            l4r4_enabled = True
            qam_hold = "hhd"
        case "r4":
            qam_button = "extra_r1"
            l4r4_enabled = True
            qam_hold = "hhd"
        case "menu":
            qam_button = "mode"
            l4r4_enabled = True
            qam_hold = "mode"
        case "disabled":
            qam_button = None
            l4r4_enabled = False
            qam_hold = "hhd"
        case _:
            qam_button = None
            l4r4_enabled = True
            qam_hold = "hhd"

    if has_touchpad:
        touch_actions = (
            conf["touchpad.controller"]
            if conf.get("touchpad.mode", None) == "controller"
            else conf["touchpad.emulation"]
        )

        multiplexer = Multiplexer(
            trigger="analog_to_discrete",
            dpad="analog_to_discrete",
            touchpad_short=touch_actions.get("short", "disabled"),
            touchpad_hold=touch_actions.get("hold", "disabled"),
            nintendo_mode=conf["nintendo_mode"].to(bool),
            qam_button=qam_button,
            emit=emit,
            params=d_params,
            # qam_multi_tap=qam_multi_tap, # supports it now
            qam_hold=qam_hold,
            startselect_chord=conf.get("main_chords", "disabled"),
        )
    else:
        multiplexer = Multiplexer(
            trigger="analog_to_discrete",
            dpad="analog_to_discrete",
            nintendo_mode=conf["nintendo_mode"].to(bool),
            qam_button=qam_button,
            emit=emit,
            params=d_params,
            # qam_multi_tap=qam_multi_tap, # supports it now
            qam_hold=qam_hold,
            startselect_chord=conf.get("main_chords", "disabled"),
        )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if motion:
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        if l4r4_enabled:
            kbd_fds = d_kbd_1.open()
            fds.extend(kbd_fds)
        else:
            kbd_fds = []
        prepare(d_xinput)
        if motion:
            start_imu = True
            try:
                if dconf.get("hrtimer", False):
                    start_imu = d_timer.open()
                if start_imu:
                    prepare(d_imu)
            except Exception as e:
                motion = False
                try:
                    d_timer.close()
                except Exception:
                    pass

                logger.error("Failed to open IMU, disabling motion support. Error:")
                # Print stacktrace
                import traceback

                traceback.print_exc()
        if has_touchpad and d_params["uses_touch"]:
            prepare(d_touch)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                # skip kbd_1 to always run it
                if f not in kbd_fds:
                    to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))
            evs.extend(d_kbd_1.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)
                d_xinput.consume(evs)

            for d in d_outs:
                d.consume(evs)

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        try:
            d_kbd_1.close(not updated.is_set())
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        try:
            d_timer.close()
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



import logging
import time
from collections import defaultdict
from typing import Sequence, cast

from hhd.controller import (
    Consumer,
    Event,
    Producer,
    DEBUG_MODE,
)
from hhd.controller.lib.uhid import UhidDevice, BUS_USB
from hhd.controller.lib.common import encode_axis, set_button
from hhd.controller.lib.ccache import ControllerCache

from .const import (
    SDCONT_VENDOR,
    SDCONT_VERSION,
    SDCONT_COUNTRY,
    SDCONT_DESCRIPTOR,
    SD_AXIS_MAP,
    SD_BTN_MAP,
    SD_SETTINGS,
)

MAX_IMU_SYNC_DELAY = 2

logger = logging.getLogger(__name__)

_cache = ControllerCache()


def trim(rep: bytes):
    if not rep:
        return rep
    idx = len(rep) - 1
    while idx > 0 and rep[idx] == 0x00:
        idx -= 1
    return rep[: idx + 1]


def pad(rep):
    return bytes(rep) + bytes([0 for _ in range(64 - len(rep))])


class SteamdeckController(Producer, Consumer):
    @staticmethod
    def close_cached():
        _cache.close()

    def __init__(
        self,
        pid,
        name,
        touchpad: bool = False,
        sync_gyro: bool = True,
    ) -> None:
        self.available = False
        self.dev = None
        self.start = 0
        self.pid = pid
        self.name = name
        self.sync_gyro = sync_gyro
        self.enable_touchpad = touchpad
        self.report = bytearray(64)
        self.i = 0
        self.last_rep = None

    def open(self) -> Sequence[int]:
        self.available = False
        self.report[0] = 0x01
        self.report[2] = 0x09
        self.report[3] = 0x40
        self.i = 0

        # Use cached controller to avoid disconnects
        cached = cast(
            SteamdeckController | None,
            _cache.get(),
        )
        self.dev = None
        if cached:
            if self.pid == cached.pid and self.name == cached.name:
                logger.warning(
                    f"Using cached controller node for Steamdeck Controller."
                )
                self.dev = cached.dev
                if self.dev and self.dev.fd:
                    self.fd = self.dev.fd
            else:
                logger.warning(f"Throwing away cached Steamdeck Controller.")
                cached.close(True, in_cache=True)
        if not self.dev:
            self.dev = UhidDevice(
                vid=SDCONT_VENDOR,
                pid=self.pid,
                bus=BUS_USB,
                version=SDCONT_VERSION,
                country=SDCONT_COUNTRY,
                name=bytes(self.name, "utf-8"),
                report_descriptor=SDCONT_DESCRIPTOR,
                unique_name=b"",
                physical_name=b"",
            )
            self.fd = self.dev.open()

        self.state: dict = defaultdict(lambda: 0)
        self.rumble = False
        self.touchpad_touch = False
        self.touchpad_left = False
        curr = time.perf_counter()
        self.start = curr
        self.touchpad_down = curr
        self.last_imu = curr
        self.imu_failed = False

        logger.info(f"Starting '{self.name}'.")
        assert self.fd
        return [self.fd]

    def close(self, exit: bool, in_cache: bool = False) -> bool:
        if not in_cache and time.perf_counter() - self.start:
            logger.warning(f"Caching Steam Controller to avoid reconnection.")
            _cache.add(self)
        elif self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if not self.fd or not self.dev or self.fd not in fds:
            return []

        # Process queued events
        out: Sequence[Event] = []
        assert self.dev
        while ev := self.dev.read_event():
            match ev["type"]:
                case "open":
                    # logger.info(f"SD OPENED")
                    pass
                case "close":
                    # logger.info(f"SD CLOSED")
                    pass
                case "start":
                    pass
                case "get_report":
                    match self.last_rep:
                        case 0x83:
                            rep = bytes(
                                [
                                    0x83,
                                    0x2D,  # 45/5=9 attrs
                                    # https://github.com/libsdl-org/SDL/blob/eed94cb0345cbf6dc9088c7bfc3d10828cb19f9d/src/joystick/hidapi/steam/controller_constants.h#L363
                                    0x01,  # ATTRIB_PRODUCT_ID
                                    0x05,
                                    0x12,
                                    0x00,
                                    0x00,
                                    0x02,  # ATTRIB_PRODUCT_REVISON
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x0A,  # ATTRIB_BOOTLOADER_BUILD_TIME
                                    0x2B,
                                    0x12,
                                    0xA9,
                                    0x62,
                                    0x04,  # ATTRIB_FIRMWARE_BUILD_TIME
                                    0xB7,
                                    0x61,
                                    0x7C,
                                    0x67,
                                    0x09,  # ATTRIB_BOARD_REVISION
                                    0x2E,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x0B,  # ATTRIB_CONNECTION_INTERVAL_IN_US
                                    0xA0,
                                    0x0F,
                                    0x00,
                                    0x00,
                                    0x0D,  # attr
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x0C,  # attr
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x0E,  # attr
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                ]
                            )
                            if not ev["rnum"]:
                                rep = bytes([0]) + rep
                        case 0xAE:
                            rep = bytes(
                                [
                                    0x00,
                                    0xAE,
                                    0x15,
                                    0x01,
                                    *[0x10 for _ in range(15)],
                                ]
                            )
                        case _:
                            rep = bytes([])
                    self.dev.send_get_report_reply(ev["id"], 0, pad(rep))
                    logger.info(
                        f"GET_REPORT: {ev}\nRESPONSE({self.last_rep:02x}): {rep.hex()}"
                    )
                case "set_report":
                    self.dev.send_set_report_reply(ev["id"], 0)
                    self.last_rep = ev["data"][3]

                    match self.last_rep:
                        case 0xEB:
                            left = int.from_bytes(
                                ev["data"][8:10], byteorder="little", signed=True
                            )
                            right = int.from_bytes(
                                ev["data"][10:12], byteorder="little", signed=True
                            )
                            out.append(
                                {
                                    "type": "rumble",
                                    "code": "main",
                                    # For some reason goes to 127
                                    "strong_magnitude": left / (2**16 - 1),
                                    "weak_magnitude": right / (2**16 - 1),
                                }
                            )
                        case 0xea:
                            # Touchpad stuff
                            pass
                        case 0x8F:
                            # logger.info(f"SD Received Haptics ({time.perf_counter()*1000:.3f}ms):\n{ev['data'].hex().rstrip(' 0')}")
                            pass
                        case 0x87:
                            if DEBUG_MODE:
                                rnum = ev["data"][4]
                                ss = []
                                for i in range(0, rnum, 3):
                                    rtype = ev["data"][5 + i]
                                    rdata = int.from_bytes(
                                        ev["data"][6 + i : 8 + i],
                                        byteorder="little",
                                        signed=False,
                                    )
                                    ss.append(
                                        f"{SD_SETTINGS[rtype] if rtype < len(SD_SETTINGS) else "UKNOWN"} ({rtype:02d}): {rdata:02x}"
                                    )
                                mlen = max(map(len, ss))
                                logger.info(
                                    f"SD Received Settings (n={rnum // 3}):{''.join(map(lambda x: '\n > ' + ' '*(mlen - len(x)) + x, ss))}"
                                )
                        case _:
                            if DEBUG_MODE:
                                logger.info(
                                    f"SD SET_REPORT({ev['rnum']:02x}:{ev['rtype']:02x}): {trim(ev['data']).hex()}"
                                )

                    # 410000eb 0901401f 0000 0000 fbfb
                    # 410000eb 0901401f ff7f ff7f fbfb
                case "output":
                    logger.info(f"SD OUTPUT")
                case _:
                    logger.warning(f"SD UKN_EVENT: {ev}")

        return out

    def consume(self, events: Sequence[Event]):
        if not self.dev:
            return

        assert self.report

        # To fix gyro to mouse in latest steam
        # only send updates when gyro sends a timestamp
        send = not self.sync_gyro
        curr = time.perf_counter()

        new_rep = bytearray(self.report)
        for ev in events:
            code = ev["code"]
            match ev["type"]:
                case "axis":
                    if not self.enable_touchpad and code.startswith("touchpad"):
                        continue
                    if code in SD_AXIS_MAP:
                        try:
                            encode_axis(new_rep, SD_AXIS_MAP[code], ev["value"])
                        except Exception:
                            logger.warning(
                                f"Encoding '{ev['code']}' with {ev['value']} overflowed."
                            )
                    match code:
                        case "gyro_ts" | "accel_ts" | "imu_ts":
                            send = True
                            self.last_imu = time.perf_counter()
                            self.last_imu_ts = ev["value"]
                case "button":
                    if not self.enable_touchpad and code.startswith("touchpad"):
                        continue
                    if code == "touchpad_touch":
                        self.touchpad_touch = ev["value"]
                        if not self.touchpad_left:
                            set_button(
                                new_rep,
                                SD_BTN_MAP["touchpad_touch"],
                                ev["value"] or self.touchpad_touch,
                            )
                    elif code == "touchpad_left":
                        set_button(
                            new_rep,
                            SD_BTN_MAP["touchpad_touch"],
                            ev["value"] or self.touchpad_touch,
                        )
                        set_button(
                            new_rep,
                            SD_BTN_MAP["touchpad_left"],
                            ev["value"],
                        )
                        encode_axis(
                            new_rep,
                            SD_AXIS_MAP["touchpad_force"],
                            ev["value"],
                        )
                        self.touchpad_left = ev["value"]
                    elif code in SD_BTN_MAP:
                        set_button(new_rep, SD_BTN_MAP[code], ev["value"])

        if not self.touchpad_touch:
            encode_axis(
                new_rep,
                SD_AXIS_MAP["touchpad_x"],
                0.5,
            )
            encode_axis(
                new_rep,
                SD_AXIS_MAP["touchpad_y"],
                0.5,
            )

        # If the IMU breaks, smoothly re-enable the controller
        failover = self.last_imu + MAX_IMU_SYNC_DELAY < curr
        if self.sync_gyro and failover and not self.imu_failed:
            self.imu_failed = True
            logger.error(
                f"IMU Did not send information for {MAX_IMU_SYNC_DELAY}s. Disabling Gyro Sync."
            )

        self.report = new_rep
        # sign_crc32_inplace(self.report, DS5_INPUT_CRC32_SEED)
        if send or failover:
            new_rep[4:8] = self.i.to_bytes(4, byteorder="little", signed=False)
            self.i = self.i + 1 if self.i < 0xFFFFFFFF else 0
            self.dev.send_input_report(self.report)
            # logger.info(self.report.hex())



import os
import shutil
import subprocess

from hhd.plugins import Context
from hhd.utils import expanduser


def find_overlay_exe(ctx: Context):
    INSTALLED_PATHS = ["hhd-ui.AppImage", "hhd-ui-dbg", "hhd-ui"]

    usr = os.environ.get("HHD_OVERLAY")
    if usr:
        if os.path.exists(usr):
            return usr
        INSTALLED_PATHS.insert(0, usr)

    # FIXME: Potential priviledge escalation attack!
    # Runs as the user in `inject_overlay`, so this should
    # not be the case. Will still be executed.
    for fn in INSTALLED_PATHS:
        local = shutil.which(fn, path=expanduser("~/.local/bin", ctx))
        if local:
            return local

    for fn in INSTALLED_PATHS:
        system = shutil.which(fn)
        if system:
            return system


def inject_overlay(fn: str, display: str, ctx: Context):
    out = subprocess.Popen(
        [fn],
        env={"HOME": expanduser("~", ctx), "DISPLAY": display, "STEAM_OVERLAY": "1"},
        text=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        preexec_fn=lambda: os.setpgrp(), # allow closing the overlay smoothly
        user=ctx.euid,
        group=ctx.egid,
    )
    return out


def launch_overlay_de(fn: str, display: str, auth: str | None, ctx: Context):
    out = subprocess.Popen(
        [fn],
        env={
            "HOME": expanduser("~", ctx),
            "XAUTHORITY": auth or "",
            "DISPLAY": display,
            "HHD_MANAGED": "1",
        },
        text=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        user=ctx.euid,
        group=ctx.egid,
    )
    return out


def get_overlay_version(fn: str, ctx: Context):
    return subprocess.run(
        [fn, "--version"],
        env={"HOME": expanduser("~", ctx)},
        text=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        user=ctx.euid,
        group=ctx.egid,
        timeout=5,
    ).stdout.strip()



import ctypes
import logging
import os
import select
import struct
import time
from fcntl import ioctl
from threading import RLock
from typing import Any, Sequence, Iterable
import stat

from evdev import InputDevice

from hhd.controller import Event as ControllerEvent
from hhd.controller.lib.ioctl import EVIOCSMASK, EVIOCGRAB
from hhd.controller.physical.evdev import B, list_evs, to_map
from hhd.controller.virtual.uinput.monkey import UInput, UInputMonkey
from hhd.utils import freeze_steam

from .const import get_touchscreen_quirk
from .x11 import is_gamescope_running
from .sdl import BTN_GAMEPAD, match_gamepad, create_joystick_guid

logger = logging.getLogger(__name__)


REFRESH_INTERVAL = 0.1
MONITOR_INTERVAL = 2
OVERLAY_BUTTON_MAP: dict[int, str] = to_map(
    {
        "a": [B("BTN_A")],
        "b": [B("BTN_B")],
        "x": [B("BTN_X")],
        "y": [B("BTN_Y")],
        "lb": [B("BTN_TL")],
        "rb": [B("BTN_TR")],
        "mode": [B("BTN_MODE")],
        "select": [B("BTN_SELECT")],
    }
)
OVERLAY_AXIS_MAP: dict[int, str] = to_map(
    {
        # Values should range from -1 to 1
        "ls_x": [B("ABS_X")],
        "ls_y": [B("ABS_Y")],
        # Hat, implemented as axis. Either -1, 0, or 1
        "hat_x": [B("ABS_HAT0X")],
        "hat_y": [B("ABS_HAT0Y")],
    }
)

CONTROLLER_WAKE_BUTTON: dict[int, str] = to_map(
    {
        "select": [B("BTN_SELECT")],
        "mode": [B("BTN_MODE")],
        "a": [B("BTN_A")],
        "b": [B("BTN_B")],
        "y": [B("BTN_Y")],
        "x": [B("BTN_X")],
    }
)

TOUCH_WAKE_AXIS: dict[int, str] = to_map(
    {
        "slot": [B("ABS_MT_SLOT")],
        "x": [B("ABS_MT_POSITION_X")],
        "y": [B("ABS_MT_POSITION_Y")],
        "id": [B("ABS_MT_TRACKING_ID")],
    }
)

KEYBOARD_WAKE_KEY: dict[int, str] = to_map(
    {
        "meta": [B("KEY_LEFTMETA")],
        "ctrl": [B("KEY_LEFTCTRL")],
        "3": [B("KEY_3")],
        "4": [B("KEY_4")],
        "armoury": [B("KEY_PROG3")],
    }
)
CUSTOM_WAKE_KEY: dict[int, str] = to_map(
    {
        "armoury": [B("KEY_PROG3")],
        "fan": [B("KEY_PROG4")],
    }
)

REPEAT_INITIAL = 0.5
REPEAT_INTERVAL = 0.2

GESTURE_TIME = 0.4
# Inspired by Ruineka's old
# gamescope gestures (pre 3.14)
GESTURE_END = 0.03
GESTURE_START = 0.02
GESTURE_TOP_RATIO = 0.33

XBOX_B_MAX_PRESS = 0.3
KBD_HOLD_DELAY = 0.5

# Cached vars
EV_ABS = B("EV_ABS")
EV_KEY = B("EV_KEY")
EV_SYN = B("EV_SYN")
HHD_DEBUG = os.environ.get("HHD_DEBUG", None)


class QamHandlerKeyboard:
    def __init__(self) -> None:
        self.uinput = None

    def _open(self):
        if self.uinput:
            return True

        args = {
            "events": {
                B("EV_KEY"): [
                    B("KEY_LEFTCTRL"),
                    B("KEY_1"),
                    B("KEY_2"),
                ]
            },
            "name": "Handheld Daemon Steam Events",
            "phys": "phys-hhd-qam",
        }
        try:
            self.uinput = UInputMonkey(**args)
            return True
        except Exception:
            try:
                self.uinput = UInput(**args)
                return True
            except Exception as e:
                pass
        return False

    def __call__(self, expanded=False) -> Any:
        if not is_gamescope_running():
            # Ctrl+1/2 do nothing outside gamescope
            return False
        if not self._open():
            return False
        if not self.uinput:
            return False

        try:
            btn = B("KEY_1") if expanded else B("KEY_2")
            self.uinput.write(B("EV_KEY"), B("KEY_LEFTCTRL"), 1)
            self.uinput.write(B("EV_KEY"), btn, 1)
            self.uinput.syn()
            time.sleep(0.3)
            self.uinput.write(B("EV_KEY"), btn, 0)
            self.uinput.write(B("EV_KEY"), B("KEY_LEFTCTRL"), 0)
            self.uinput.syn()
            return True
        except Exception as e:
            logger.error(f"Could not send keyboard event. Error:\n{e}")
            return False

    def screenshot(self) -> bool:
        if not self._open():
            return False
        if not self.uinput:
            return False

        try:
            btn = B("KEY_F12")
            self.uinput.write(B("EV_KEY"), btn, 1)
            self.uinput.syn()
            time.sleep(0.1)
            self.uinput.write(B("EV_KEY"), btn, 0)
            self.uinput.syn()
            return True
        except Exception as e:
            logger.error(f"Could not send screenshot event. Error:\n{e}")
            return False

    def close(self):
        if self.uinput:
            self.uinput.close()
            self.uinput = None


def grab_buttons(fd: int, typ: int, btns: Iterable[int] | None):
    if btns:
        b_len = max((max(btns) >> 3) + 1, 8)
        # Align to long
        if b_len % 8 != 0:
            b_len += 8 - (b_len % 8)
        mask = bytearray(b_len)
        for b in btns:
            mask[b >> 3] |= 1 << (b & 0x07)
    else:
        mask = bytes([])
        b_len = 0

    c_mask = ctypes.create_string_buffer(bytes(mask))
    data = struct.pack("@ I I L", typ, b_len, ctypes.addressof(c_mask))
    # Before
    # print(bytes(mask).hex())
    ioctl(fd, EVIOCSMASK, data)
    # After
    # ioctl(fd, EVIOCGMASK, data)
    # print(bytes(mask).hex())


_is_ally = None


def is_ally():
    global _is_ally

    if _is_ally is not None:
        return _is_ally

    try:
        with open("/sys/devices/virtual/dmi/id/product_name", "r") as f:
            product_name = f.read().strip().lower()
            _is_ally = "ally" in product_name
            return _is_ally
    except Exception:
        return False


def has_touchscreen():
    try:
        for dev in list_evs():
            keys = dev.get("byte", {}).get("key", bytes())
            major = B("BTN_TOUCH") >> 3
            minor = B("BTN_TOUCH") & 0x07
            has_touch = len(keys) > major and keys[major] & (1 << minor)
            major = B("BTN_TOOL_FINGER") >> 3
            minor = B("BTN_TOOL_FINGER") & 0x07
            has_finger = len(keys) > major and keys[major] & (1 << minor)
            if has_touch and not has_finger:
                return True
    except Exception:
        return True
    return False


def find_devices(
    current: dict[str, Any] = {},
    keyboard: bool = True,
    controllers: bool = True,
    touchscreens: bool = True,
    sdl_mappings: Any | None = None,
):
    out = {}
    for name, dev in list_evs(True).items():
        if name in current:
            continue

        # Skip HHD devices
        if "hhd" in dev.get("phys", "") or (
            # Allow bluetooth controllers that contain uhid and phys, while
            # blocking hhd devices that contain uhid but not phys
            "uhid" in dev.get("sysfs", "")
            and not dev.get("phys", "")
        ):
            continue

        # Skip Steam virtual devices
        # Vendor=28de Product=11ff
        if dev.get("vendor", 0) == 0x28DE and dev.get("product", 0) == 0x11FF:
            continue

        abs = dev.get("byte", {}).get("abs", bytes())
        keys = dev.get("byte", {}).get("key", bytes())

        # Touchscreen is complicated. Should have BTN_TOUCH but not BTN_TOOL_FINGER
        is_touchscreen = touchscreens
        major = B("BTN_TOUCH") >> 3
        minor = B("BTN_TOUCH") & 0x07
        if len(keys) <= major or not keys[major] & (1 << minor):
            is_touchscreen = False
        major = B("BTN_TOOL_FINGER") >> 3
        minor = B("BTN_TOOL_FINGER") & 0x07
        if len(keys) > major and keys[major] & (1 << minor):
            is_touchscreen = False

        for cap in TOUCH_WAKE_AXIS:
            major = cap >> 3
            minor = cap & 0x07
            if len(abs) <= major or not abs[major] & (1 << minor):
                is_touchscreen = False
                break

        is_controller = controllers
        for cap in CONTROLLER_WAKE_BUTTON:
            major = cap >> 3
            minor = cap & 0x07
            if len(keys) <= major or not keys[major] & (1 << minor):
                is_controller = False
                break

        # Use SDL mappings
        cap = BTN_GAMEPAD
        major = cap >> 3
        minor = cap & 0x07
        guid = create_joystick_guid(dev)
        if (
            sdl_mappings
            and controllers
            and len(keys) > major
            and keys[major] & (1 << minor)
        ):
            sdl_info = match_gamepad(dev, sdl_mappings)
            if sdl_info:
                is_controller = False
            else:
                sdl_info = {}
        else:
            sdl_info = {}

        # Avoid laptop keyboards, as they emit left meta on power button hold
        # FIXME: will prevent using laptop keyboards to bring up the menu
        is_keyboard = keyboard and not dev.get("name", "").startswith("AT Translated")
        for cap in KEYBOARD_WAKE_KEY:
            major = cap >> 3
            minor = cap & 0x07
            if len(keys) <= major or not keys[major] & (1 << minor):
                is_keyboard = False
                break

        is_custom = False
        for cap, v in CUSTOM_WAKE_KEY.items():
            major = cap >> 3
            minor = cap & 0x07

            # Only bind armoury to Asus WMI hotkeys
            if v == "armoury" and (
                "Asus WMI hotkeys" not in dev.get("name", "") or is_ally()
            ):
                continue

            if v == "fan" and ("asus" not in dev.get("name", "").lower() or is_ally()):
                continue

            if len(keys) > major and keys[major] & (1 << minor):
                is_custom = True
                break

        if is_touchscreen or is_controller or is_keyboard or is_custom or sdl_info:
            out[name] = {
                "is_touchscreen": is_touchscreen,
                "is_controller": is_controller,
                "is_keyboard": is_keyboard,
                "is_custom": is_custom,
                "pretty": dev.get("name", ""),
                "hash": dev.get("hash", ""),
                "vid": dev.get("vendor", 0),
                "pid": dev.get("product", 0),
                "guid": guid.hex(),
                "sdl_info": sdl_info,
            }

    return out


def process_touch(emit, state, ev, val):
    # Check if the gesture should be kept
    invalidated = False
    if ev == "slot" and val:
        # Second finger, remove the gesture
        invalidated = True
    elif ev == "id" and val == -1:
        # Finger removed, remove the gesture
        invalidated = True
        # This is the only time we remove the
        # start_time as well, so gestures can resume.
        state["start_time"] = 0

    if invalidated:
        state["start_x"] = 0
        state["start_y"] = 0
        state["last_x"] = 0
        state["last_y"] = 0
        state["grab"] = False
        return

    start_time = state.get("start_time", 0)
    curr = time.time()
    if start_time and curr - start_time > GESTURE_TIME:
        # User took too long, stop processing gestures
        # until finger is released
        return

    # After this point, we only use coordinates
    if ev not in ("x", "y"):
        return

    # Swap names around to avoid
    # Confusion with portrait displays
    if state["portrait"]:
        if ev == "x":
            ev = "y"
            max_ev = state["max_x"]
        elif ev == "y":
            ev = "x"
            max_ev = state["max_y"]
    else:
        max_ev = state[f"max_{ev}"]

    if state["flip_x"] and ev == "x":
        val = max_ev - val
    if state["flip_y"] and ev == "y":
        val = max_ev - val

    if not start_time:
        state["start_time"] = curr

    # Save old values
    v = val / max_ev
    state[f"start_{ev}"] = state[f"start_{ev}"] if state.get(f"start_{ev}", 0) else v
    state[f"last_{ev}"] = v

    # Begin handler
    start_x = state.get("start_x", 0)
    start_y = state.get("start_y", 0)
    last_x = state.get("last_x", 0)
    last_y = state.get("last_y", 0)

    if not start_x or not start_y:
        return
    if not last_x or not last_y:
        return

    if (
        start_x < GESTURE_START
        or start_x > 1 - GESTURE_START
        or start_y > 1 - GESTURE_START
    ):
        state["grab"] = True

    # logger.info(
    #     f"{start_x:.2f}:{start_y:.2f} -> {last_x:.2f}:{last_y:.2f} = ({dx:5.2f}, {dy:5.2f})"
    # )

    handled = False
    if start_x < GESTURE_START and last_x > GESTURE_END:
        semi = "top" if start_y < GESTURE_TOP_RATIO else "bottom"
        if emit:
            emit({"type": "special", "event": f"swipe_right_{semi}"})
        handled = True
    elif start_x > 1 - GESTURE_START and last_x < 1 - GESTURE_END:
        semi = "top" if start_y < GESTURE_TOP_RATIO else "bottom"
        if emit:
            emit({"type": "special", "event": f"swipe_left_{semi}"})
        handled = True
    elif start_y > 1 - GESTURE_START and last_y < 1 - GESTURE_END:
        if emit:
            emit({"type": "special", "event": "swipe_bottom"})
        handled = True
    elif start_y < GESTURE_START and last_y > GESTURE_END:
        if emit:
            emit({"type": "special", "event": "swipe_top"})
        handled = True

    if handled:
        state["start_x"] = 0
        state["start_y"] = 0
        state["last_x"] = 0
        state["last_y"] = 0
        state["grab"] = False


def process_custom(emit, ev, val):
    if ev == "armoury" and val and emit:
        emit({"type": "special", "event": "custom_armoury"})
    if ev == "fan" and val and emit:
        emit({"type": "special", "event": "custom_fan"})


def process_kbd(emit, state, ev, val):
    if ev == "ctrl":
        state["ctrl"] = val
        return
    if ev == "3" and val and state.get("ctrl", 0):
        if emit:
            emit({"type": "special", "event": "kbd_ctrl_3"})
    if ev == "4" and val and state.get("ctrl", 0):
        if emit:
            emit({"type": "special", "event": "kbd_ctrl_4"})

    # Skip repeats
    if val == 2:
        return

    if not ev == "meta":
        return

    pressed_n = state.get("pressed_n", 0)

    curr = time.time()
    if val:
        state["pressed_n"] = pressed_n + 1
        state["last_pressed"] = curr
    else:
        if pressed_n:
            emit({"type": "special", "event": "kbd_meta_press"})
        state["last_pressed"] = 0


def refresh_kbd(emit, state):
    pressed_n = state.get("pressed_n", 0)
    last_pressed = state.get("last_pressed", 0)
    # last_release = state.get("last_release", 0)
    curr = time.time()

    if pressed_n and last_pressed and curr - last_pressed > KBD_HOLD_DELAY:
        if emit:
            emit({"type": "special", "event": "kbd_meta_hold"})
        state["pressed_n"] = 0
        state["last_pressed"] = 0


def process_ctrl(emit, state, ev, val, allow_select=True):
    # Here, we capture the shortcut xbox+b
    # This is a shortcut that is used by steam
    # but only when it is held, so we can use
    # it if its a shortpress
    if ev == "mode":
        state["mode"] = val
        return
    if ev == "select":
        state["select"] = val
        return

    if ev != "a" and ev != "b" and ev != "y" and ev != "x":
        return

    # Mode needs to be pressed
    if not state.get("mode", None) and not (state.get("select", None) and allow_select):
        return

    if val:
        state[ev] = time.time()
    else:
        if state.get(ev, None) and time.time() - state[ev] < XBOX_B_MAX_PRESS:
            logger.info(f"Xbox+{ev} pressed")
            if emit:
                emit(
                    {
                        "type": "special",
                        "event": f"xbox_{ev}",
                        "data": {"uniq": state.get("uniq", None)},
                    }
                )
        state[ev] = None


def process_events(emit, dev, evs, allow_select=True):
    # Some nice logging to make things easier
    if HHD_DEBUG:
        log = ""
        if dev["is_touchscreen"]:
            for ev in evs:
                if ev.type != EV_ABS:
                    continue
                code = ev.code
                if code not in TOUCH_WAKE_AXIS:
                    continue
                log += f"\n - {TOUCH_WAKE_AXIS[code]}: {ev.value}"
        elif dev["is_controller"]:
            for ev in evs:
                if ev.type != EV_KEY:
                    continue
                code = ev.code
                if code not in CONTROLLER_WAKE_BUTTON:
                    continue
                log += f"\n - {CONTROLLER_WAKE_BUTTON[code]}: {ev.value}"
        elif dev["is_keyboard"]:
            for ev in evs:
                if ev.type != EV_KEY:
                    continue
                code = ev.code
                if code not in KEYBOARD_WAKE_KEY:
                    continue
                if ev.value == 2:
                    continue
                log += f"\n - {KEYBOARD_WAKE_KEY[code]}: {ev.value}"
        elif dev["is_custom"]:
            for ev in evs:
                if ev.type != EV_KEY:
                    continue
                code = ev.code
                if code not in CUSTOM_WAKE_KEY:
                    continue
                if ev.value == 2:
                    continue
                log += f"\n - {CUSTOM_WAKE_KEY[code]}: {ev.value}"
        if log:
            logger.info(f"'{dev['pretty']}':{log}")

    for ev in evs:
        # The evs list is SYN, however, for gestures do we really care?
        # We can also do some ugly prefiltering here, so that the
        # inner functions are prettier
        if ev.type == EV_SYN:
            continue

        if dev["is_touchscreen"] and ev.type == EV_ABS and ev.code in TOUCH_WAKE_AXIS:
            process_touch(emit, dev["state_touch"], TOUCH_WAKE_AXIS[ev.code], ev.value)

        if (
            dev["is_controller"]
            and ev.type == EV_KEY
            and ev.code in CONTROLLER_WAKE_BUTTON
        ):
            process_ctrl(
                emit, dev["state_ctrl"], CONTROLLER_WAKE_BUTTON[ev.code], ev.value, allow_select=allow_select
            )

        if (
            dev["sdl_info"]
            and ev.type == EV_KEY
            and ev.code in dev["sdl_info"]["buttons"]
        ):
            process_ctrl(
                emit, dev["state_ctrl"], dev["sdl_info"]["buttons"][ev.code], ev.value
            )

        if dev["is_keyboard"] and ev.type == EV_KEY and ev.code in KEYBOARD_WAKE_KEY:
            process_kbd(emit, dev["state_kbd"], KEYBOARD_WAKE_KEY[ev.code], ev.value)

        if dev["is_custom"] and ev.type == EV_KEY and ev.code in CUSTOM_WAKE_KEY:
            process_custom(emit, CUSTOM_WAKE_KEY[ev.code], ev.value)


def refresh_events(emit, dev):
    # if dev["is_touchscreen"]:
    #     refresh_touch(emit, dev["state_touch"])
    # if dev["is_controller"]:
    #     refresh_ctrl(emit, dev["state_ctrl"])
    if dev["is_keyboard"]:
        refresh_kbd(emit, dev["state_kbd"])


def intercept_devices(devs, activate: bool):
    failed = []
    for id, dev in devs.items():
        if not dev["is_controller"] and not dev["sdl_info"]:
            continue
        d = dev["dev"]
        if activate:
            try:
                grab_buttons(
                    d.fd,
                    B("EV_KEY"),
                    [
                        *KEYBOARD_WAKE_KEY,
                        *CUSTOM_WAKE_KEY,
                        *dev["sdl_info"].get("wake_buttons", CONTROLLER_WAKE_BUTTON),
                        *dev["sdl_info"].get("overlay_buttons", OVERLAY_BUTTON_MAP),
                    ],
                )
                grab_buttons(
                    d.fd,
                    B("EV_ABS"),
                    dev["sdl_info"].get("overlay_axes", OVERLAY_AXIS_MAP),
                )

                if not dev.get("grabbed", False):
                    ioctl(d.fd, EVIOCGRAB, 2)
                    dev["grabbed"] = True
                logger.info(f" - '{dev['pretty']}'")
            except Exception:
                logger.warning(f" - Failed: '{dev['pretty']}'")
                failed.append((id, dev))
        else:
            try:
                if dev.get("grabbed", False):
                    d.ungrab()
                    dev["grabbed"] = False

                grab_buttons(
                    d.fd,
                    B("EV_KEY"),
                    [
                        *dev["sdl_info"].get("wake_buttons", CONTROLLER_WAKE_BUTTON),
                        *KEYBOARD_WAKE_KEY,
                        *CUSTOM_WAKE_KEY,
                    ],
                )
                grab_buttons(d.fd, B("EV_ABS"), {})
                logger.info(f" - '{dev['pretty']}'")
            except Exception:
                logger.warning(f" - Failed: '{dev['pretty']}'")
                failed.append((id, dev))
    return failed


def intercept_events(emit, intercept_num, cid, abs, sdl, evs):
    if not emit:
        return

    out = []
    for ev in evs:
        if ev.type == EV_SYN:
            continue

        overlay_buttons = sdl.get("buttons", OVERLAY_BUTTON_MAP)
        overlay_axes = sdl.get("axes", OVERLAY_AXIS_MAP)

        if ev.type == EV_KEY and ev.code in overlay_buttons:
            out.append(
                {
                    "type": "button",
                    "code": overlay_buttons[ev.code],
                    "value": ev.value,
                }
            )
        elif ev.type == EV_ABS and ev.code in overlay_axes:
            code = overlay_axes[ev.code]

            smin = abs[ev.code]["min"]
            smax = abs[ev.code]["max"]
            v = min(
                1,
                max(
                    0,
                    (ev.value - smin) / (smax - smin),
                ),
            )

            if isinstance(code, dict):
                dtype = code["type"]
                info = code
                code = info.get("code", "")
            else:
                dtype = "axis"
                info = {}

            if info.get("flip", False):
                v = 1 - v

            if "trigger" not in code:
                v = v * 2 - 1

            if dtype == "axis":
                out.append(
                    {
                        "type": "axis",
                        "code": code,
                        "value": v,
                    }
                )
            elif dtype == "button":
                if v > AXIS_LIMIT:
                    v = 1
                else:
                    v = 0
                out.append(
                    {
                        "type": "button",
                        "code": code,
                        "value": v,
                    }
                )
            elif dtype == "hat":
                vu = 0
                vd = 0
                if v > AXIS_LIMIT:
                    vu = 1
                elif v < -AXIS_LIMIT:
                    vd = 1

                out.append(
                    {
                        "type": "button",
                        "code": info["up_code"],
                        "value": vu,
                    }
                )
                out.append(
                    {
                        "type": "button",
                        "code": info["down_code"],
                        "value": vd,
                    }
                )

    emit.intercept(cid + intercept_num, out)


def device_shortcut_loop(
    emit=None,
    should_exit=None,
    init=True,
    keyboard: bool = True,
    controllers: bool = True,
    touchscreens: bool = True,
    disable_touchscreens: bool = False,
    touch_correction: dict | None = None,
    sdl_mappings: Any | None = None,
    allow_select: bool = True,
):
    blacklist = set()
    last_check = 0
    intercept = False
    intercept_num = 0
    steam_frozen = False
    devs = {}
    while not should_exit or not should_exit.is_set():
        if devs:
            # Wait for events
            try:
                r, _, _ = select.select(
                    [d["dev"].fd for d in devs.values()], [], [], REFRESH_INTERVAL
                )
            except Exception:
                pass
        elif not init:
            # If no devices, wait for a bit
            # Except on first run
            time.sleep(MONITOR_INTERVAL)
        init = False

        # Process events
        should_intercept = emit and emit.should_intercept()
        if any(dev["is_controller"] or dev["sdl_info"] for dev in devs.values()):
            failed = []
            if not intercept and should_intercept:
                intercept = True
                intercept_num += 1
                logger.info("Intercepting other controllers:")
                failed = intercept_devices(devs, True)

                if emit and not steam_frozen and len(devs) > len(failed):
                    steam_frozen = freeze_steam(True, emit.ctx)
                    if steam_frozen:
                        logger.info("Froze Steam (to avoid HID device dual input)")
            elif intercept and not should_intercept:
                intercept = False
                logger.info("Stopping intercepting other controllers:")
                failed = intercept_devices(devs, False)

            for id, f in failed:
                blacklist.add(f["hash"])
                try:
                    del devs[id]
                except Exception:
                    pass

        if emit and steam_frozen and not should_intercept:
            # Give time for the B event to be lost
            time.sleep(0.25)
            logger.info("Unfreezing Steam")
            freeze_steam(False, emit.ctx)
            steam_frozen = False

        for name, dev in list(devs.items()):
            d = dev["dev"]
            refresh_events(emit, dev)
            if not d.fd in r:
                # Run interception so that holding button repeats work
                if should_intercept and (dev["is_controller"] or dev["sdl_info"]):
                    intercept_events(
                        emit,
                        intercept_num,
                        dev["hash"],
                        dev["abs_info"],
                        dev["sdl_info"],
                        [],
                    )
                continue

            try:
                if (dev["is_controller"] or dev["sdl_info"]) and os.stat(
                    name
                ).st_mode & stat.S_IRGRP == 0:
                    logger.info(f"Removing hidden device: '{dev['pretty']}'")
                    blacklist.add(cand["hash"])
                    del devs[name]
                    try:
                        d.close()
                    except Exception:
                        pass
                    continue

                e = list(d.read())
                # print(e)
                process_events(emit, dev, e, allow_select=allow_select)
                if should_intercept and (dev["is_controller"] or dev["sdl_info"]):
                    intercept_events(
                        emit,
                        intercept_num,
                        dev["hash"],
                        dev["abs_info"],
                        dev["sdl_info"],
                        e,
                    )
            except Exception as e:
                logger.error(
                    f"Device '{dev['pretty']}' has error. Removing. Error:\n{e}"
                )
                blacklist.add(dev["hash"])
                del devs[name]
                try:
                    d.close()
                except Exception:
                    pass

        # Avoid spamming proc
        curr = time.time()
        if curr - last_check < MONITOR_INTERVAL:
            continue

        # Add new devices
        log = ""
        for name, cand in find_devices(
            devs,
            keyboard=keyboard,
            touchscreens=touchscreens or disable_touchscreens,
            controllers=controllers,
            sdl_mappings=sdl_mappings,
        ).items():
            if cand["hash"] in blacklist:
                continue

            try:
                if os.stat(name).st_mode & stat.S_IRGRP == 0:
                    continue

                dev = InputDevice(name)

                if cand["is_touchscreen"] and disable_touchscreens:
                    # Grab touchscreen if requested
                    dev.grab()

                # Add event filters to avoid CPU use
                # Do controllers and keyboards together as buttons do not consume much
                if (
                    cand["sdl_info"]
                    or cand["is_controller"]
                    or cand["is_keyboard"]
                    or cand["is_custom"]
                ):
                    grab_buttons(
                        dev.fd,
                        B("EV_KEY"),
                        [
                            *cand["sdl_info"].get(
                                "wake_buttons", CONTROLLER_WAKE_BUTTON
                            ),
                            *KEYBOARD_WAKE_KEY,
                            *CUSTOM_WAKE_KEY,
                        ],
                    )
                else:
                    grab_buttons(dev.fd, B("EV_KEY"), {})

                # Abs events
                # Touchscreen, joystick, etc. We only care about touchscreens
                if cand["is_touchscreen"]:
                    grab_buttons(dev.fd, B("EV_ABS"), TOUCH_WAKE_AXIS)
                else:
                    grab_buttons(dev.fd, B("EV_ABS"), {})

                # Rel events are not used
                # They contain e.g., scroll events, mouse movements
                grab_buttons(dev.fd, B("EV_REL"), {})
                # MSC Events are not used
                # They contain e.g., scan codes
                grab_buttons(dev.fd, B("EV_MSC"), {})

                devs[name] = {
                    "dev": dev,
                    **cand,
                    "state_touch": {},
                    "state_ctrl": {},
                    "state_kbd": {},
                }
                caps = []
                if cand["is_touchscreen"]:
                    max_x = dev.absinfo(B("ABS_MT_POSITION_X")).max
                    max_y = dev.absinfo(B("ABS_MT_POSITION_Y")).max

                    # Default quirks
                    portrait = max_x < max_y
                    flip_x = not portrait  # just the way it is
                    flip_y = False

                    quirk, pretty = get_touchscreen_quirk(
                        vid=dev.info.vendor, pid=dev.info.product
                    )
                    if touch_correction:
                        portrait = touch_correction.get("portrait", False)
                        flip_x = touch_correction.get("flip_x", False)
                        flip_y = touch_correction.get("flip_y", False)
                        caps.append(
                            f"Touchscreen[manual, portrait={portrait}, x={flip_x}, y={flip_y}]"
                        )
                    elif quirk:
                        portrait = quirk.portrait
                        flip_x = quirk.flip_x
                        flip_y = quirk.flip_y
                        caps.append(f"Touchscreen[{pretty}]")
                    else:
                        caps.append(
                            f"Touchscreen[auto, portrait={portrait}, x={flip_x}, y={flip_y}]"
                        )

                    devs[name]["state_touch"].update(
                        {
                            "max_x": max_x,
                            "max_y": max_y,
                            "portrait": portrait,
                            "flip_x": flip_x,
                            "flip_y": flip_y,
                        }
                    )

                # Common
                if cand["is_controller"] or cand["sdl_info"]:
                    cap = dev.capabilities(verbose=False, absinfo=True)
                    devs[name]["abs_info"] = {k: {"min": v.min, "max": v.max} for k, v in cap.get(EV_ABS, [])}  # type: ignore
                    devs[name]["state_ctrl"].update({"uniq": dev.uniq})
                # Controller
                if cand["is_controller"]:
                    caps.append(f"Controller[ukn={cand['guid']}]")
                if cand["sdl_info"]:
                    pretty = cand["sdl_info"].get("name", "Unknown")
                    guid = cand["sdl_info"].get("guid", None)
                    if guid:
                        guid = guid.hex()
                    else:
                        guid = None
                    caps.append(f'SDL[{guid}="{pretty}"]')

                if cand["is_keyboard"]:
                    caps.append("Keyboard")
                if cand["is_custom"]:
                    caps.append("Custom")
                log += f"\n - '{cand['pretty']}' [{cand['vid']:04x}:{cand['pid']:04x}] ({', '.join(caps)})"
            except Exception as e:
                logger.error(f"Failed to open device '{cand['pretty']}'. Error:\n{e}")
                blacklist.add(cand["hash"])

        if log:
            logger.info(f"Found new shortcut devices:{log}")


AXIS_LIMIT = 0.5


class OverlayWriter:
    def __init__(self, stdout, mute: bool = True) -> None:
        self.state = {}
        self.stdout = stdout
        self._write_lock = RLock()

        if mute:
            # We support intercepting all controllers now
            self.write("cmd:mute\n")

    def _call(self, cid: int, evs: Sequence[ControllerEvent]):
        if not cid in self.state:
            self.state[cid] = {}

        # Debounce changed events
        curr = time.perf_counter()
        changed = []
        for ev in evs:
            match ev["type"]:
                case "axis":
                    code = ev["code"]
                    act1 = act2 = val1 = val2 = None
                    if code == "ls_x" or code == "hat_x":
                        act1 = "right"
                        act2 = "left"
                        if ev["value"] > AXIS_LIMIT:
                            val1 = True
                            val2 = False
                        elif ev["value"] < -AXIS_LIMIT:
                            val1 = False
                            val2 = True
                        else:
                            val1 = False
                            val2 = False
                    elif code == "ls_y" or code == "hat_y":
                        act1 = "down"
                        act2 = "up"
                        if ev["value"] > AXIS_LIMIT:
                            val1 = True
                            val2 = False
                        elif ev["value"] < -AXIS_LIMIT:
                            val1 = False
                            val2 = True
                        else:
                            val1 = False
                            val2 = False

                    if act1 and act2:
                        for code, val in ((act1, val1), (act2, val2)):
                            if (
                                code not in self.state[cid]
                                or bool(self.state[cid][code]) != val
                            ):
                                changed.append((code, val))
                                self.state[cid][code] = (
                                    curr + REPEAT_INITIAL if val else None
                                )
                case "button":
                    code = ev["code"]
                    if code not in (
                        "a",
                        "b",
                        "x",
                        "y",
                        "rb",
                        "lb",
                        "mode",
                        "select",
                        "dpad_up",
                        "dpad_down",
                        "dpad_left",
                        "dpad_right",
                    ):
                        continue
                    if "dpad_" in code:
                        code = code.replace("dpad_", "")
                    val = ev["value"]
                    if (
                        code not in self.state[cid]
                        or bool(self.state[cid][code]) != val
                    ):
                        changed.append((code, val))
                        self.state[cid][code] = curr + REPEAT_INITIAL if val else None

        # Ignore guide combos
        if self.state[cid].get("mode", None) or self.state[cid].get("select", None):
            return

        # Allow holds
        for btn, val in list(self.state[cid].items()):
            if not val:
                continue
            if val < curr:
                changed.append((btn, True))
                self.state[cid][btn] = curr + REPEAT_INTERVAL

        # Process changed events
        cmds = ""
        for code, val in changed:
            if code == "b":
                # This is tricky, if we do not wait for release,
                # input will leak to steam and into the game.
                # So for b specifically, wait for up
                if not val:
                    cmds += f"action:b\n"
            elif val:
                cmds += f"action:{code}\n"
            elif code == "x":
                cmds += f"action:x_up\n"

        # Write them out
        if cmds:
            self.write(cmds)

    def __call__(self, cid: int, evs: Sequence[ControllerEvent]):
        with self._write_lock:
            return self._call(cid, evs)

    def write(self, cmds: str):
        try:
            with self._write_lock:
                self.stdout.write(cmds)
                self.stdout.flush()
        except Exception:
            pass

    def reset(self):
        with self._write_lock:
            self.state = {}



import logging
import time
from typing import Sequence

try:
    from evdev.uinput import UInput
except ImportError:
    from evdev import UInput # type: ignore[no-redef]

from hhd.controller.base import Consumer, Event, Producer, can_read
from hhd.controller.const import Axis, Button

from .const import *
from .monkey import UInputMonkey
from hhd.controller.lib.ccache import ControllerCache

logger = logging.getLogger(__name__)

_cache = ControllerCache()
_cache_motions = ControllerCache()
_cache_volume = ControllerCache()

MAX_IMU_SYNC_DELAY = 2


class UInputDevice(Consumer, Producer):
    @staticmethod
    def close_cached():
        _cache.close()
        _cache_motions.close()

    @staticmethod
    def close_volume_cached():
        _cache_volume.close()

    def __init__(
        self,
        capabilities=GAMEPAD_CAPABILITIES,
        btn_map: dict[Button, int] = GAMEPAD_BUTTON_MAP,
        axis_map: dict[Axis, AX] = GAMEPAD_AXIS_MAP,
        vid: int = HHD_VID,
        pid: int = HHD_PID_GAMEPAD,
        bus: int = 0x03,
        name: str = "Handheld Daemon Controller",
        phys: str = "phys-hhd-gamepad",
        output_imu_timestamps: str | bool = False,
        output_timestamps: bool = False,
        input_props: Sequence[int] = [],
        ignore_cmds: bool = False,
        uniq: str | None = None,
        version: int = 1,
        cache: bool = False,
        motions_device: bool = False,
        volume_keyboard: bool = False,
        sync_gyro: bool = False,
    ) -> None:
        self.capabilities = capabilities
        self.btn_map = btn_map
        self.axis_map = axis_map
        self.dev = None
        self.name = name
        self.vid = vid
        self.pid = pid
        self.bus = bus
        self.phys = phys
        self.uniq = uniq
        self.output_imu_timestamps = output_imu_timestamps
        self.output_timestamps = output_timestamps
        self.last_imu_ts = 0
        self.start = 0
        self.input_props = input_props
        self.ignore_cmds = ignore_cmds
        self.version = version
        self.cache = cache
        self.motions_device = motions_device
        self.volume_keyboard = volume_keyboard
        self.sync_gyro = sync_gyro
        self.imu_failed = False
        self.last_imu = 0
        self.wrote = False
        if volume_keyboard:
            self.cache = True

        self.rumble: Event | None = None

    def open(self) -> Sequence[int]:
        logger.info(f"Opening virtual device '{self.name}'.")
        self.dev = None

        if self.cache:
            if self.motions_device:
                name = "left motions device"
                cache = _cache_motions.get()
            elif self.volume_keyboard:
                name = "volume keyboard"
                cache = _cache_volume.get()
            else:
                name = "controller"
                cache = _cache.get()

            cached = cast(UInputDevice | None, cache)
            if cached:
                if (
                    self.capabilities == cached.capabilities
                    and self.name == cached.name
                    and self.vid == cached.vid
                    and self.pid == cached.pid
                    and self.bus == cached.bus
                    and self.phys == cached.phys
                    and self.input_props == cached.input_props
                    and self.uniq == cached.uniq
                ):
                    logger.warning(f"Using cached controller node for {name}.")
                    self.dev = cached.dev
                else:
                    cached.close(True, in_cache=True)

        if not self.dev:
            try:
                self.dev = UInputMonkey(
                    events=self.capabilities,
                    name=self.name,
                    vendor=self.vid,
                    product=self.pid,
                    version=self.version,
                    bustype=self.bus,
                    phys=self.phys,
                    input_props=self.input_props,
                    uniq=self.uniq,
                )
            except Exception as e:
                logger.error(
                    f"Monkey patch probably failed. Could not create evdev device with uniq:\n{e}"
                )
                self.dev = UInput(
                    events=self.capabilities,
                    name=self.name,
                    vendor=self.vid,
                    product=self.pid,
                    bustype=self.bus,
                    version=self.version,
                    phys=self.phys,
                    input_props=self.input_props,
                )

        self.touchpad_aspect = 1
        self.touch_id = 1
        self.fd = self.dev.fd
        self.start = time.perf_counter()
        self.last_imu = time.perf_counter()
        self.imu_failed = False
        self.wrote = False

        if self.ignore_cmds:
            # Do not wake up if we ignore to save utilization
            # When the output contains a timestamp, it is fed back to evdev
            # causing double wake-ups.
            return []
        return [self.fd]

    def close(self, exit: bool, in_cache: bool = False) -> bool:
        if not in_cache and self.cache:
            if self.motions_device:
                name = "left motions device"
                _cache_motions.add(self)
            elif self.volume_keyboard:
                name = "volume keyboard"
                _cache_volume.add(self)
            else:
                name = "controller"
                _cache.add(self)
            logger.warning(f"Caching {name} to avoid reconnection.")
        elif self.dev:
            self.dev.close()
            self.dev = None
            self.input = None
            self.fd = None
        return True

    def consume(self, events: Sequence[Event]):
        if not self.dev:
            return

        should_syn = not self.sync_gyro
        wrote = {}
        ts = 0
        for ev in reversed(events):
            key = (ev["type"], ev["code"])
            if key in wrote:
                # skip duplicate events that were caused due to a delay
                # only keep the last button value by iterating reversed
                continue
            match ev["type"]:
                case "axis":
                    if not should_syn and "imu_ts" in ev["code"]:
                        self.imu_failed = False
                        self.last_imu = time.perf_counter()
                        should_syn = True

                    if ev["code"] in self.axis_map:
                        ax = self.axis_map[ev["code"]]
                        if ev["code"] == "touchpad_x":
                            val = int(
                                self.touchpad_aspect
                                * (ax.scale * ev["value"] + ax.offset)
                            )
                        else:
                            val = int(ax.scale * ev["value"] + ax.offset)
                        if ax.bounds:
                            val = min(max(val, ax.bounds[0]), ax.bounds[1])
                        self.dev.write(B("EV_ABS"), ax.id, val)
                        wrote[key] = val

                        if ev["code"] == "touchpad_x":
                            self.dev.write(B("EV_ABS"), B("ABS_MT_POSITION_X"), val)
                        elif ev["code"] == "touchpad_y":
                            self.dev.write(B("EV_ABS"), B("ABS_MT_POSITION_Y"), val)

                    elif (
                        self.output_imu_timestamps is True
                        and ev["code"]
                        in (
                            "accel_ts",
                            "gyro_ts",
                            "imu_ts",
                        )
                    ) or ev["code"] == self.output_imu_timestamps:
                        # We have timestamps with ns accuracy.
                        # Evdev expects us accuracy
                        self.last_imu_ts = ev["value"]
                        ts = (ev["value"] // 1000) % (2**31)
                        self.dev.write(B("EV_MSC"), B("MSC_TIMESTAMP"), ts)
                        wrote[key] = ts
                case "button":
                    if ev["code"] in self.btn_map:
                        if ev["code"] == "touchpad_touch":
                            self.dev.write(
                                B("EV_ABS"),
                                B("ABS_MT_TRACKING_ID"),
                                self.touch_id if ev["value"] else -1,
                            )
                            self.dev.write(
                                B("EV_KEY"),
                                B("BTN_TOOL_FINGER"),
                                1 if ev["value"] else 0,
                            )
                            self.touch_id += 1
                            if self.touch_id > 500:
                                self.touch_id = 1
                        self.dev.write(
                            B("EV_KEY"),
                            self.btn_map[ev["code"]],
                            1 if ev["value"] else 0,
                        )
                        wrote[key] = ev["value"]

                case "configuration":
                    if ev["code"] == "touchpad_aspect_ratio":
                        self.touchpad_aspect = float(ev["value"])

        if wrote and self.output_timestamps:
            # We have timestamps with ns accuracy.
            # Evdev expects us accuracy
            ts = (time.perf_counter_ns() // 1000) % (2**31)
            self.dev.write(B("EV_MSC"), B("MSC_TIMESTAMP"), ts)

        if self.sync_gyro:
            curr = time.perf_counter()
            if curr - self.last_imu > MAX_IMU_SYNC_DELAY and not self.imu_failed:
                self.imu_failed = True
                logger.error(
                    f"IMU Did not send information for {MAX_IMU_SYNC_DELAY}s. Disabling Gyro Sync."
                )

        self.wrote = self.wrote or bool(wrote)
        if (
            self.wrote
            and (should_syn or not self.sync_gyro or self.imu_failed)
            and (not self.output_imu_timestamps or ts)
        ):
            self.dev.syn()
            self.wrote = False

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.ignore_cmds or not self.fd or not self.fd in fds or not self.dev:
            return []

        out: Sequence[Event] = []

        while can_read(self.fd):
            for ev in self.dev.read():
                if ev.type == B("EV_MSC") and ev.code == B("MSC_TIMESTAMP"):
                    # Skip timestamp feedback
                    # TODO: Figure out why it feedbacks
                    pass
                elif ev.type == B("EV_UINPUT"):
                    if ev.code == B("UI_FF_UPLOAD"):
                        # Keep uploaded effect to apply on input
                        upload = self.dev.begin_upload(ev.value)
                        if upload.effect.type == B("FF_RUMBLE"):
                            data = upload.effect.u.ff_rumble_effect

                            self.rumble = {
                                "type": "rumble",
                                "code": "main",
                                "weak_magnitude": data.weak_magnitude / 0xFFFF,
                                "strong_magnitude": data.strong_magnitude / 0xFFFF,
                            }
                        self.dev.end_upload(upload)
                    elif ev.code == B("UI_FF_ERASE"):
                        # Ignore erase events
                        erase = self.dev.begin_erase(ev.value)
                        erase.retval = 0
                        self.dev.end_erase(erase)
                elif ev.type == B("EV_FF") and ev.value:
                    if self.rumble:
                        out.append(self.rumble)
                    else:
                        logger.warning(
                            f"Rumble requested but a rumble effect has not been uploaded.\n{ev}"
                        )
                elif ev.type == B("EV_FF") and not ev.value:
                    out.append(
                        {
                            "type": "rumble",
                            "code": "main",
                            "weak_magnitude": 0,
                            "strong_magnitude": 0,
                        }
                    )
                else:
                    logger.info(f"Controller ev received unhandled event:\n{ev}")

        return out



from threading import Event, Thread
from typing import Any, Sequence

from hhd.controller.lib.hid import enumerate_unique
from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    get_gyro_config,
    get_outputs_config,
    get_touchpad_config,
    load_relative_yaml,
)
from hhd.plugins.settings import HHDSettings
from hhd.utils import get_distro_color, hsb_to_rgb

from .const import (
    GPD_WIN_4_8840U_MAPPINGS,
    GPD_WIN_DEFAULT_MAPPINGS,
    GPD_WIN_MAX_2_2023_MAPPINGS,
)

GPD_CONFS = {
    "G1618-03": {  # Old model, has no gyro/touchpad
        "name": "GPD Win 3",
        "touchpad": False,
        "hrtimer": False,
    },
    "G1618-04": {
        "name": "GPD Win 4",
        "hrtimer": True,
        "wincontrols": True,
        "rgb": True,
        "combo": "menu",
        "chord": "select",
    },
    "G1617-01": {
        "name": "GPD Win Mini",
        "touchpad": True,
        "wincontrols": True,
    },
    "G1617-02": {
        "name": "GPD Win Mini (2025)",
        "touchpad": True,
        "wincontrols": False,
    },
    "G1619-04": {
        "name": "GPD Win Max 2 (04)",
        "hrtimer": True,
        "touchpad": True,
        "mapping": GPD_WIN_MAX_2_2023_MAPPINGS,
        "wincontrols": True,
    },
    "G1619-05": {
        "name": "GPD Win Max 2 (05)",
        "hrtimer": True,
        "touchpad": True,
        "mapping": GPD_WIN_MAX_2_2023_MAPPINGS,
        "wincontrols": True,
    },
}


def get_default_config(product_name: str):
    return {
        "name": product_name,
        "hrtimer": True,
        "untested": True,
    }


class GpdWinControllersPlugin(HHDPlugin):
    name = "gpd_win_controllers"
    priority = 18
    log = "gpdw"

    def __init__(self, dmi: str, dconf: dict) -> None:
        self.t = None
        self.should_exit = None
        self.updated = Event()
        self.started = False
        self.t = None

        self.dmi = dmi
        self.dconf = dconf
        self.name = f"gpd_win_controllers@'{dconf.get('name', 'ukn')}'"

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

    def settings(self) -> HHDSettings:
        base = {"controllers": {"gpd_win": load_relative_yaml("controllers.yml")}}
        base["controllers"]["gpd_win"]["children"]["controller_mode"].update(
            get_outputs_config(
                can_disable=True,
                has_leds=False,
                start_disabled=self.dconf.get("untested", False),
            )
        )

        # Tweak defaults for l4r4menu and main_chords
        base["controllers"]["gpd_win"]["children"]["l4r4"]["default"] = self.dconf.get(
            "combo", "r4"
        )
        base["controllers"]["gpd_win"]["children"]["main_chords"]["default"] = (
            self.dconf.get("chord", "disabled")
        )

        if self.dconf.get("touchpad", False):
            base["controllers"]["gpd_win"]["children"][
                "touchpad"
            ] = get_touchpad_config()
        else:
            del base["controllers"]["gpd_win"]["children"]["touchpad"]

        base["controllers"]["gpd_win"]["children"]["imu_axis"] = get_gyro_config(
            self.dconf.get("mapping", GPD_WIN_DEFAULT_MAPPINGS)
        )
        return base

    def update(self, conf: Config):
        new_conf = conf["controllers.gpd_win"]
        if new_conf == self.prev:
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        self.updated.set()
        self.start(self.prev)

    def start(self, conf):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = Event()
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                self.dconf,
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None


class GpdWinControlsPlugin(HHDPlugin):
    name = "gpd_wincontrols"
    priority = 18
    log = "gpdc"

    def __init__(self, dmi: str, dconf: dict) -> None:
        self.dmi = dmi
        self.dconf = dconf
        self.name = f"gpd_wincontrols@'{dconf.get('name', 'ukn')}'"

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context

    def settings(self) -> HHDSettings:
        base = {"wincontrols": {"wincontrols": load_relative_yaml("wincontrols.yml")}}

        if self.dconf.get("rgb", False):
            hue = get_distro_color()
            base["wincontrols"]["wincontrols"]["children"]["leds"]["modes"]["solid"][
                "children"
            ]["hue"]["default"] = hue
            base["wincontrols"]["wincontrols"]["children"]["leds"]["modes"]["pulse"][
                "children"
            ]["hue"]["default"] = hue
        else:
            del base["wincontrols"]["wincontrols"]["children"]["leds"]

        return base

    def update(self, conf: Config):
        if not conf.get_action(f"wincontrols.wincontrols.apply"):
            return

        from .wincontrols import (
            BACKBUTTONS_DEFAULT,
            BACKBUTTONS_HHD,
            BUTTONS_DEFAULT,
            BUTTONS_PHAWX,
            BUTTONS_TRIGGERS_DEFAULT,
            BUTTONS_TRIGGERS_STEAMOS,
            update_config,
        )

        c = conf["wincontrols.wincontrols"]
        vibration = c.get("vibration", "off")

        buttons = {}
        delays = {}
        deadzones = {}

        match c.get("mouse_mode", "unchanged"):
            case "mouse":
                buttons.update(BUTTONS_DEFAULT)
            case "wasd":
                buttons.update(BUTTONS_PHAWX)

        match c.get("mouse_mode_triggers", "unchanged"):
            case "gpd":
                buttons.update(BUTTONS_TRIGGERS_DEFAULT)
            case "steamos":
                buttons.update(BUTTONS_TRIGGERS_STEAMOS)

        match c.get("l4r4", "unchanged"):
            case "hhd":
                buttons.update(BACKBUTTONS_HHD["buttons"])
                delays.update(BACKBUTTONS_HHD["delays"])
            case "default":
                buttons.update(BACKBUTTONS_DEFAULT["buttons"])
                delays.update(BACKBUTTONS_DEFAULT["delays"])

        if c.get("deadzones.mode", "unchanged") == "custom":
            deadzones.update(c.get("deadzones.custom", {}))

        rgb_mode = "off"
        rgb_color = (0, 0, 0)
        match c.get("leds.mode", "disabled"):
            case "disabled":
                rgb_mode = "off"
            case "solid":
                rgb_mode = "constant"
                hue = c.get("leds.solid.hue", 0)
                rgb_color = tuple(hsb_to_rgb(hue, 100, 100))
            case "pulse":
                rgb_mode = "breathed"
                hue = c.get("leds.pulse.hue", 0)
                rgb_color = tuple(hsb_to_rgb(hue, 100, 100))
            case "rainbow":
                rgb_mode = "rotated"

        try:
            conf["wincontrols.wincontrols.fwver"] = update_config(
                buttons,
                delays,
                rumble=vibration,
                rgb_mode=rgb_mode,
                rgb_color=rgb_color,  # type: ignore
                deadzones=deadzones,
            )
        except Exception as e:
            conf["wincontrols.wincontrols.status"] = f"{e}"
            conf["wincontrols.wincontrols.fwver"] = f""


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    try:
        # Match just product number, should be enough for now
        with open("/sys/devices/virtual/dmi/id/product_name") as f:
            dmi = f.read().strip()
            dconf = GPD_CONFS.get(dmi, None)

            if dmi == "G1618-04":
                with open("/proc/cpuinfo") as f:
                    cpuinfo = f.read().strip()
                # 8840U has a different gyro mapping
                if "AMD Ryzen 7 8840U" in cpuinfo:
                    dconf = dict(GPD_CONFS["G1618-04"])
                    dconf["name"] = "GPD Win 4 (8840U)"
                    dconf["mapping"] = GPD_WIN_4_8840U_MAPPINGS

            if dconf:
                base: list[HHDPlugin] = [GpdWinControllersPlugin(dmi, dconf)]
                if dconf.get("wincontrols", False):
                    base.append(GpdWinControlsPlugin(dmi, dconf))
                return base

        with open("/sys/devices/virtual/dmi/id/sys_vendor") as f:
            vendor = f.read().strip().lower()
        if vendor == "gpd":
            return [GpdWinControllersPlugin(dmi, get_default_config(dmi))]
    except Exception:
        pass

    return []



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map
from hhd.controller.physical.hidraw import AM, BM, CM

GOS_INTERFACE_BTN_ESSENTIALS: dict[Button, BM] = {
    # Misc
    "mode": BM((0 << 3) + 7),
    "share": BM((0 << 3) + 6),
    # Back buttons
    "extra_l1": BM((2 << 3) + 7),
    "extra_r1": BM((2 << 3) + 6),
}


GOS_INTERFACE_BTN_MAP: dict[Button, BM] = {
    # Misc
    "mode": BM((0 << 3) + 7),
    "share": BM((0 << 3) + 6),
    # Sticks
    "ls": BM((0 << 3) + 5),
    "rs": BM((0 << 3) + 4),
    # D-PAD
    "dpad_up": BM((0 << 3) + 3),
    "dpad_down": BM((0 << 3) + 2),
    "dpad_left": BM((0 << 3) + 1),
    "dpad_right": BM((0 << 3) + 0),
    # Thumbpad
    "a": BM((1 << 3) + 7),
    "b": BM((1 << 3) + 6),
    "x": BM((1 << 3) + 5),
    "y": BM((1 << 3) + 4),
    # Bumpers
    "lb": BM((1 << 3) + 3),
    "lt": BM((1 << 3) + 2),
    "rb": BM((1 << 3) + 1),
    "rt": BM((1 << 3) + 0),
    # Back buttons
    "extra_l1": BM((2 << 3) + 7),
    "extra_r1": BM((2 << 3) + 6),
    # Select
    "start": BM((2 << 3) + 0),
    "select": BM((2 << 3) + 1),
}


GOS_INTERFACE_AXIS_MAP: dict[Axis, AM] = {
    "ls_x": AM(4 << 3, "m8"),
    "ls_y": AM(5 << 3, "m8"),
    "rs_x": AM(6 << 3, "m8"),
    "rs_y": AM(7 << 3, "m8"),
    "rt": AM(12 << 3, "u8"),
    "lt": AM(13 << 3, "u8"),
    # # Controller IMU
    "accel_x": AM(14 << 3, "i16", scale=-0.00212, order="little"),
    "accel_z": AM(16 << 3, "i16", scale=-0.00212, order="little"),
    "accel_y": AM(18 << 3, "i16", scale=-0.00212, order="little"),
    "gyro_x": AM(20 << 3, "i16", scale=-0.0005325, order="little"),
    "gyro_z": AM(22 << 3, "i16", scale=0.0005325, order="little"),
    "gyro_y": AM(24 << 3, "i16", scale=-0.0005325, order="little"),
}

GOS_TOUCHPAD_BUTTON_MAP: dict[int, Button] = to_map(
    {
        "touchpad_touch": [B("BTN_TOOL_FINGER")],  # also BTN_TOUCH
        "touchpad_left": [B("BTN_LEFT")],
    }
)

GOS_TOUCHPAD_AXIS_MAP: dict[int, Axis] = to_map(
    {
        "touchpad_x": [B("ABS_X")],  # also ABS_MT_POSITION_X
        "touchpad_y": [B("ABS_Y")],  # also ABS_MT_POSITION_Y
    }
)

GOS_TOUCHPAD_AXIS_MAP_HID: dict[int | None, dict[Axis, AM]] = {
    None: {
        "touchpad_x": AM(2 << 3, "i16", scale=1/400),
        "touchpad_y": AM(4 << 3, "i16", scale=1/400),
    }
}
GOS_TOUCHPAD_BUTTON_MAP_HID: dict[int | None, dict[Button, BM]] = {
    None: {
        "touchpad_touch": BM((8 << 3) + 7),
        "touchpad_left": BM((9 << 3) + 7),
    }
}



import logging
import os
import select
import time
from threading import Event, Thread
from typing import Literal, Sequence, TypedDict, cast

from hhd.controller.base import RgbMode
from hhd.controller.lib.hid import Device as HIDDevice
from hhd.controller.lib.hid import enumerate_unique
from hhd.i18n import _
from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    get_outputs_config,
    load_relative_yaml,
)
from hhd.plugins.settings import HHDSettings
from hhd.utils import get_distro_color, hsb_to_rgb

logger = logging.getLogger(__name__)

SEARCH_INTERVAL = 10

AURA_CONFIGS_LIGHTBAR = {
    "solid": ["color"],
    "pulse": ["color", "speedd"],
    "duality": ["dual", "speedd"],
    "cycle": ["speedd", "direction"],
    "rainbow": ["speedd"],
    "strobe": ["color", "speedd"],
}

AURA_CONFIGS_USB = {
    "solid": ["color"],
    "pulse": ["color", "speedd"],
    "duality": ["dual", "speedd"],
    # "cycle": ["speedd", "direction"],
    "rainbow": ["speedd"],
    "strobe": ["color", "speedd"],
}

ASUS_VID = 0x0B05
BRIGHTNESS_MAP = ["disabled", "low", "medium", "high"]
AURA_APPLICATIONS = [0xFF310076, 0xFF310079, 0xFF310080]

AURA_CONFIGS = {
    # Z13 Lightbar
    0x18C6: (_("Lightbar"), AURA_CONFIGS_LIGHTBAR),
    # ROG Keyboards (Z13 incl.)
    0x1A30: (_("Keyboard"), AURA_CONFIGS_USB),
}

AURA_CONFGIS_WMI = {
    "solid": ["color"],
}

WMI_LOCATION = "/sys/class/leds/asus::kbd_backlight/brightness"
WMI_NOTIFICATION = "/sys/class/leds/asus::kbd_backlight/brightness_hw_changed"

RGB_INPUT_ID = 0x5A
RGB_AURA_ID = 0x5D

RGB_HANDSHAKE = lambda key: bytes(
    [
        key,
        0x41,
        0x53,
        0x55,
        0x53,
        0x20,
        0x54,
        0x65,
        0x63,
        0x68,
        0x2E,
        0x49,
        0x6E,
        0x63,
        0x2E,
    ]
)
RGB_INIT_OTHER = lambda key: [
    bytes([key, 0xB9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]),
    bytes([key, 0x05, 0x20, 0x31, 0, 0x1A]),
]
DYN_RGB_INIT = lambda key: buf([key, 0xC0, 0x03, 0x01])
RGB_APPLY = lambda key: buf([key, 0xB4])
RGB_SET = lambda key: buf([key, 0xB5, 0, 0, 0])
RGB_APPLY_DELAY = 1
RGB_TDP_DELAY = 1.5


def init_rgb_dev(key: int, dev: HIDDevice):
    # Send handshake command
    init_cmd = RGB_HANDSHAKE(key)
    dev.send_feature_report(init_cmd)

    # Device replies with the same start
    rec = dev.get_feature_report(key, 64)
    for a1, a2 in zip(init_cmd, rec):
        assert a1 == a2, f"Init command mismatch:\n" f"{init_cmd.hex()} != {rec.hex()}"

    # Maybe not required
    # # More initialization commands
    # for cmd in RGB_INIT_OTHER(key):
    #     dev.send_feature_report(cmd)


class AuraDevice(TypedDict):
    vid: int
    pid: int
    application: int
    fn: str
    name: str
    cfg_name: str
    init: bool
    disabled: bool
    modes: dict[str, list[str]]
    dev: HIDDevice
    last_mode: RgbMode | None


def buf(x):
    return bytes(x) + bytes(64 - len(x))


def monitor_brightness(emit, should_exit: Event):
    if not os.path.exists(WMI_NOTIFICATION):
        logger.warning("WMI notification file does not exist: %s", WMI_NOTIFICATION)
        return

    with open(WMI_NOTIFICATION, "r") as f:
        p = select.poll()
        p.register(f.fileno(), select.POLLPRI)
        while not should_exit.is_set():
            # Use a time limit for exit to be possible
            v = p.poll(2000)
            if not v:
                continue
            try:
                f.read()
                f.seek(0)
            except OSError as e:
                if e.errno == 61:
                    continue
                raise
            emit({"type": "special", "event": "brightness_changed"})


def get_aura_devices(
    existing: dict[str, AuraDevice] = {},
) -> tuple[dict[str, AuraDevice], bool]:
    out = {}

    found = set()
    for d in enumerate_unique(vid=ASUS_VID):
        application = d.get("usage_page", 0x0000) << 16 | d.get("usage", 0x0000)
        if d["path"] in existing:
            ref = existing[d["path"]]
            if (
                d["vendor_id"] != ref["vid"]
                or d["product_id"] != ref["pid"]
                or application != ref["application"]
            ):
                # Device changed, do not add it to found
                continue
            # Skip already known devices
            found.add(d["path"])
            continue
        if application not in AURA_APPLICATIONS:
            continue
        if d["product_id"] not in AURA_CONFIGS:
            continue

        name, modes = AURA_CONFIGS[d["product_id"]]
        cfg_name = f"{d['vendor_id']:04x}_{d['product_id']:04x}"

        try:
            dev = HIDDevice(path=d["path"])
        except Exception as e:
            logger.warning(
                "Failed to open Aura device %s (%04x:%04x): %s",
                d["path"],
                d["vendor_id"],
                d["product_id"],
                e,
            )
            continue

        out[d["path"]] = AuraDevice(
            name=name,
            cfg_name=cfg_name,
            vid=d["vendor_id"],
            pid=d["product_id"],
            application=application,
            fn=d["path"],
            init=True,
            disabled=False,
            modes=modes,
            dev=dev,
            last_mode=None,
        )
        logger.info(
            "Found Aura device %s (%s, %04x:%04x) with modes:\n%s",
            name,
            d["path"].decode("utf-8"),
            d["vendor_id"],
            d["product_id"],
            list(modes),
        )

    updated = False
    for k in list(existing.keys()):
        if k not in found:
            logger.info(
                "Removing Aura device %s (%s, %04x:%04x) from known devices",
                existing[k]["fn"],
                existing[k]["name"],
                existing[k]["vid"],
                existing[k]["pid"],
            )
            del existing[k]
            updated = True

    updated |= bool(out)
    return (
        dict(sorted({**existing, **out}.items(), key=lambda x: x[1]["cfg_name"])),
        updated,
    )


def rgb_command(
    mode: RgbMode,
    direction,
    speed: str,
    red: int,
    green: int,
    blue: int,
    o_red: int,
    o_green: int,
    o_blue: int,
):
    c_direction = 0x00
    set_speed = True

    match mode:
        case "solid":
            # Static
            c_mode = 0x00
            set_speed = False
        case "pulse":
            # Strobing
            # c_mode = 0x0A
            # Spiral is agressive
            # Use breathing instead
            # Breathing
            c_mode = 0x01
            o_red = 0
            o_green = 0
            o_blue = 0
        case "strobe":
            # Strobing
            c_mode = 0x0A
        case "rainbow":
            # Color cycle
            c_mode = 0x02
        case "spiral" | "cycle":
            # Rainbow
            c_mode = 0x03
            red = 0
            green = 0
            blue = 0
            if direction == "left":
                c_direction = 0x01
        case "duality":
            # Breathing
            c_mode = 0x01
        # case "direct":
        #     # Direct/Aura
        #     c_mode = 0xFF
        # Should be used for dualsense emulation/ambilight stuffs
        case _:
            c_mode = 0x00

    c_speed = 0xE1
    if set_speed:
        match speed:
            case "low":
                c_speed = 0xE1
            case "medium":
                c_speed = 0xEB
            case _:  # "high"
                c_speed = 0xF5

    c_zone = 0x00
    return buf(
        [
            RGB_AURA_ID,
            0xB3,
            c_zone,  # zone
            c_mode,  # mode
            red,
            green,
            blue,
            c_speed if mode != "solid" else 0x00,
            c_direction,
            0x00,  # breathing
            o_red,  # these only affect the breathing mode
            o_green,
            o_blue,
            0,
            0,
            0,
            0,
        ]
    )


def get_aura_mode_cmd(cfg, dev: AuraDevice):
    # Get event info
    mode = cfg["mode"].to(str)
    if mode in cfg:
        info = cast(dict, cfg[mode].conf)
    else:
        info = {}
    if not dev["modes"] or mode not in dev["modes"]:
        return None, mode, False

    red = green = blue = red2 = green2 = blue2 = 0
    speedd = "medium"
    direction = "left"
    color2_set = False
    log = f"Setting Aura {dev['name']} ({dev['vid']:04x}:{dev['pid']:04x}) to mode '{mode}'"
    always_init = True
    for cap in dev["modes"][mode]:
        match cap:
            case "color":
                red, green, blue = hsb_to_rgb(
                    info["hue"],
                    info["saturation"],
                    100,
                )
                log += f" with color: {red:3d}, {green:3d}, {blue:3d}"
                always_init = False
            case "dual":
                red, green, blue = hsb_to_rgb(
                    info["hue"],
                    info["saturation"],
                    100,
                )
                red2, green2, blue2 = hsb_to_rgb(
                    info["hue2"],
                    info["saturation"],
                    100,
                )
                color2_set = True
                log += f" with colors: {red:3d}, {green:3d}, {blue:3d} and {red2:3d}, {green2:3d}, {blue2:3d}"
                always_init = False
            case "speedd":
                log += f", speed: {info['speedd']}"
                speedd = cast(Literal["low", "medium", "high"], info["speedd"])
            case "direction":
                log += f", direction: {info['direction']}"
                direction = cast(Literal["left", "right"], info["direction"])

    # logger.info(log)
    return (
        rgb_command(
            mode,
            direction,
            speedd,
            red,
            green,
            blue,
            red2 if color2_set else 0,
            green2 if color2_set else 0,
            blue2 if color2_set else 0,
        ),
        mode,
        always_init,
    )


def set_aura_brightness(
    brightness: str, devices: dict[str, AuraDevice], has_wmi: bool = False
) -> bool:
    if not devices:
        return False

    match brightness:
        case "high":
            c = 0x03
        case "medium":
            c = 0x02
        case "low":
            c = 0x01
        case _:
            c = 0x00

    if os.path.exists(WMI_LOCATION) and has_wmi:
        try:
            with open(WMI_LOCATION, "w") as f:
                f.write(str(c))
            logger.info("Set Aura brightness to %s", brightness)
            return False
        except Exception as e:
            logger.error("Failed to set WMI brightness: %s", e)
            return False

    error = False
    for dev in devices.values():
        if dev["disabled"]:
            continue
        try:
            # Set brightness on the device
            dev["dev"].send_feature_report(buf([RGB_INPUT_ID, 0xBA, 0xC5, 0xC4, c]))
        except Exception as e:
            logger.error(
                "Failed to set brightness for Aura device %s (%04x:%04x): %s",
                dev["fn"],
                dev["vid"],
                dev["pid"],
                e,
            )
            dev["disabled"] = True
            try:
                dev["dev"].close()
            except Exception:
                pass
            error = True
    return error


def get_aura_power_cmd(power) -> bytes:
    boot = power.get("boot", False)
    awake = power.get("awake", False)
    sleep = power.get("sleep", False)
    shutdown = power.get("shutdown", False)

    keyb = 0
    bar = 0
    lid_rear = 0

    if boot:
        keyb |= (1 << 0) | (1 << 1)
        bar |= 1 << 1
        lid_rear |= (1 << 0) | (1 << 4)

    if awake:
        keyb |= (1 << 2) | (1 << 3)
        bar |= (1 << 0) | (1 << 2)
        lid_rear |= (1 << 1) | (1 << 5)

    if sleep:
        keyb |= (1 << 4) | (1 << 5)
        bar |= 1 << 3
        lid_rear |= (1 << 2) | (1 << 6)

    if shutdown:
        keyb |= (1 << 6) | (1 << 7)
        bar |= 1 << 4
        lid_rear |= (1 << 3) | (1 << 7)

    return bytes(
        [
            RGB_AURA_ID,
            0xBD,
            0x01,
            keyb,
            bar,
            lid_rear,
            lid_rear,
            0xFF,
        ]
    )


def set_aura_power(power, devs) -> bool:
    error = False
    cmd = get_aura_power_cmd(power)

    for dev in devs.values():
        if dev["disabled"]:
            continue
        try:
            # Set brightness on the device
            dev["dev"].send_feature_report(cmd)
        except Exception as e:
            logger.error(
                "Failed to set brightness for Aura device %s (%04x:%04x): %s",
                dev["fn"],
                dev["vid"],
                dev["pid"],
                e,
            )
            dev["disabled"] = True
            try:
                dev["dev"].close()
            except Exception:
                pass
            error = True
    return error


class AuraPlugin(HHDPlugin):
    name = "aura_rgb_controller"
    priority = 18
    log = "aura"

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.queue_apply = {}
        self.tdp_changes = False
        self.context = context
        self.enabled = False
        self.loaded_devices = set()
        self.devices = {}
        self.prev_cfg = None
        self.has_wmi = os.path.exists(WMI_LOCATION)
        self.init_brightness = not self.has_wmi

        # Set a sane default color
        self.dc = get_distro_color()

        self.t = None
        self.should_exit = Event()

    def settings(self) -> HHDSettings:
        # If RGB support is disabled
        # return enable option only
        base = load_relative_yaml("settings.yml")
        self.loaded_devices = set()
        if not self.enabled:
            del base["rgb"]
            return base

        if not self.devices:
            return {}

        modes = load_relative_yaml("../../plugins/rgb/modes.yml")
        capabilities = load_relative_yaml("../../plugins/rgb/capabilities.yml")

        cfgs = {}
        for d in self.devices.values():
            if d["disabled"]:
                continue
            supported = {}
            for mode, caps in d["modes"].items():
                if mode in modes:
                    m = modes[mode]
                    m["children"] = {}
                    for cap in caps:
                        # We use four step brightness for the keyboard button to work
                        m["children"].update(
                            {
                                k: dict(v)
                                for k, v in capabilities[cap].items()
                                if k != "brightness"
                            }
                        )
                        if cap == "color":
                            m["children"]["hue"]["default"] = self.dc
                    for c in m["children"].values():
                        c["tags"] = sorted(set(c.get("tags", []) + m.get("tags", [])))
                    supported[mode] = m

            # Add supported modes
            cfg = {
                "type": "mode",
                "title": d["name"],
                "modes": supported,
            }

            # Set a sane default mode
            for default in ("solid", "pulse"):
                if default in supported:
                    cfg["default"] = default
                    break
            else:
                # fallback to any supported mode to have persistence in the mode
                cfg["default"] = next(iter(supported))

            cfgs[d["cfg_name"]] = cfg
            self.loaded_devices.add(d["cfg_name"])

        base_settings = base["rgb"]["aura"]["children"]
        base["rgb"]["aura"]["children"] = {
            "brightness": base_settings["brightness"],
            **cfgs,
            "power": base_settings["power"],
        }

        if self.has_wmi:
            # If the driver is loaded, we will get the brightness from the os
            del base["rgb"]["aura"]["children"]["brightness"]["default"]

        return base

    def update(self, conf: Config):
        enabled_prev = self.enabled
        self.enabled = conf.get("hhd.settings.aura", False)

        if enabled_prev != self.enabled:
            self.emit({"type": "settings"})

        if not self.enabled:
            return

        curr_t = time.perf_counter()
        error = False
        if self.prev_cfg is not None and "rgb.aura" in conf:
            curr = conf["rgb.aura"]
            power_settings = curr.get("power", None)
            if power_settings is not None:
                self.tdp_changes = power_settings.get("tdp_changes", False)

            # Set per device settings
            for d in self.devices.values():
                k = d["cfg_name"]
                if k not in self.loaded_devices:
                    continue
                if d["disabled"]:
                    continue
                if k not in curr:
                    continue

                data = curr[k]
                init = d["init"]
                queued = k in self.queue_apply and self.queue_apply[k] < curr_t
                if queued:
                    del self.queue_apply[k]
                if (
                    k in self.prev_cfg
                    and data == self.prev_cfg[k]
                    and not init
                    and not queued
                ):
                    continue
                d["init"] = False

                cmd, new_mode, always_init = get_aura_mode_cmd(data, d)
                if cmd is None:
                    continue
                chanded_mode = new_mode != d["last_mode"]

                try:
                    if init:
                        init_rgb_dev(RGB_AURA_ID, d["dev"])
                        if power_settings is not None:
                            # Set power settings on init
                            d["dev"].send_feature_report(
                                get_aura_power_cmd(power_settings),
                            )

                        # Needs time to initialize, get it next cycle
                        self.queue_apply[k] = curr_t + RGB_APPLY_DELAY
                    else:
                        d["dev"].send_feature_report(cmd)
                        d["last_mode"] = new_mode

                        if chanded_mode or queued or always_init or init:
                            d["dev"].send_feature_report(RGB_SET(RGB_AURA_ID))
                            d["dev"].send_feature_report(RGB_APPLY(RGB_AURA_ID))
                        else:
                            self.queue_apply[k] = curr_t + RGB_APPLY_DELAY

                except Exception as e:
                    logger.error(
                        f"Failed to set mode to {d['name']} ({d['fn']}, {d['vid']:04x}_{d['pid']:04x}), removing:\n{e}",
                    )
                    d["disabled"] = True
                    try:
                        d["dev"].close()
                    except Exception:
                        pass
                    error = True
                    continue

            # Set common settings (allow to init)
            brightness = curr.get("brightness", None)
            if brightness is not None and (
                brightness != self.prev_cfg.get("brightness", None)
                or self.init_brightness
            ):
                error |= set_aura_brightness(brightness, self.devices, self.has_wmi)
                self.init_brightness = False

            if (
                power_settings is not None
                and self.prev_cfg.get("power", None)
                and (power_settings != self.prev_cfg.get("power", None))
            ):
                error |= set_aura_power(power_settings, self.devices)

        self.devices, updated = get_aura_devices(self.devices)
        if updated:
            logger.info("Found %d Aura devices", len(self.devices))
            self.emit({"type": "settings"})
        if error:
            self.emit({"type": "settings"})

        if self.has_wmi:
            if self.t is None:
                self.should_exit.clear()
                self.t = Thread(
                    target=monitor_brightness,
                    args=(self.emit, self.should_exit),
                    name="Aura WMI brightness monitor",
                )
                self.t.start()

            with open(WMI_LOCATION, "r") as f:
                brightness = int(f.read().strip())
            br_str = BRIGHTNESS_MAP[min(3, brightness)]
            conf["rgb.aura.brightness"] = br_str
        
        if self.loaded_devices and "rgb.aura" in conf:
            self.prev_cfg = conf["rgb.aura"]

    def close(self):
        if self.t is not None:
            self.should_exit.set()
            self.t.join()
            self.t = None

        for d in self.devices.values():
            try:
                d["dev"].close()
            except Exception as e:
                logger.warning(
                    "Failed to close Aura device %s (%04x:%04x): %s",
                    d["fn"],
                    d["vid"],
                    d["pid"],
                    e,
                )

        self.devices = {}

    def notify(self, events):
        if not self.tdp_changes:
            return

        color = None
        for ev in events:
            if ev["type"] == "special":
                match ev["event"]:
                    case "tdp_cycle_quiet":
                        color = (0, 0, 255)
                    case "tdp_cycle_balanced":
                        color = (255, 255, 255)
                    case "tdp_cycle_performance":
                        color = (255, 0, 0)
                    case "tdp_cycle_custom":
                        color = (157, 0, 255)
                    # case "wakeup":
                    #     for d in self.devices.values():
                    #         if d["disabled"]:
                    #             continue
                    #         d['init'] = True
                    case _:
                        color = None

        if color is None:
            return

        error = False
        for d in self.devices.values():
            if d["disabled"]:
                continue
            try:
                cmd = rgb_command(
                    "solid",
                    "left",
                    "medium",
                    *color,
                    0,  # o_red
                    0,  # o_green
                    0,  # o_blue
                )
                d["dev"].send_feature_report(cmd)
                d["dev"].send_feature_report(RGB_SET(RGB_AURA_ID))
                self.queue_apply[d["cfg_name"]] = (
                    time.perf_counter() + RGB_TDP_DELAY
                )
            except Exception as e:
                logger.error(
                    f"Failed to set TDP cycle color on {d['name']} ({d['vid']:04x}:{d['pid']:04x}): {e}",
                )
                d["disabled"] = True
                error = True
                try:
                    d["dev"].close()
                except Exception:
                    pass

        if error:
            self.emit({"type": "settings"})

def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product name
    # if a device exists here its officially supported
    with open("/sys/devices/virtual/dmi/id/sys_vendor") as f:
        vendor = f.read().strip()

    # Match just product number, should be enough for now
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        # Different variants of the ally can have an additional _RC71L or not
        dmi = f.read().strip()

    if vendor == "ASUSTeK COMPUTER INC." and "ROG Ally" not in dmi:
        return [AuraPlugin()]

    return []



import logging
import time

from hhd.controller.lib.hid import Device, enumerate_unique

logger = logging.getLogger(__name__)

ACTION_MAP = {
    "none": 0x00,
    # Standard HID codes
    # Letters
    "a": 0x04,
    "b": 0x05,
    "c": 0x06,
    "d": 0x07,
    "e": 0x08,
    "f": 0x09,
    "g": 0x0A,
    "h": 0x0B,
    "i": 0x0C,
    "j": 0x0D,
    "k": 0x0E,
    "l": 0x0F,
    "m": 0x10,
    "n": 0x11,
    "o": 0x12,
    "p": 0x13,
    "q": 0x14,
    "r": 0x15,
    "s": 0x16,
    "t": 0x17,
    "u": 0x18,
    "v": 0x19,
    "w": 0x1A,
    "x": 0x1B,
    "y": 0x1C,
    "z": 0x1D,
    # Numbers
    "1": 0x1E,
    "2": 0x1F,
    "3": 0x20,
    "4": 0x21,
    "5": 0x22,
    "6": 0x23,
    "7": 0x24,
    "8": 0x25,
    "9": 0x26,
    "0": 0x27,
    # Special characters
    "enter": 0x28,
    "escape": 0x29,
    "backspace": 0x2A,
    "tab": 0x2B,
    "space": 0x2C,
    "minus": 0x2D,
    "equal": 0x2E,
    "leftbrace": 0x2F,
    "rightbrace": 0x30,
    "backslash": 0x31,
    "hashtilde": 0x32,
    "semicolon": 0x33,
    "apostrophe": 0x34,
    "grave": 0x35,
    "comma": 0x36,
    "dot": 0x37,
    "slash": 0x38,
    "capslock": 0x39,
    "f1": 0x3A,
    "f2": 0x3B,
    "f3": 0x3C,
    "f4": 0x3D,
    "f5": 0x3E,
    "f6": 0x3F,
    "f7": 0x40,
    "f8": 0x41,
    "f9": 0x42,
    "f10": 0x43,
    "f11": 0x44,
    "f12": 0x45,
    "sysrq": 0x46,
    "scrolllock": 0x47,
    "pause": 0x48,
    "insert": 0x49,
    "home": 0x4A,
    "pageup": 0x4B,
    "delete": 0x4C,
    "end": 0x4D,
    "pagedown": 0x4E,
    "right": 0x4F,
    "left": 0x50,
    "down": 0x51,
    "up": 0x52,
    "numlock": 0x53,
    "kpslash": 0x54,
    "kpasterisk": 0x55,
    "kpminus": 0x56,
    "kpplus": 0x57,
    "kpenter": 0x58,
    "kp1": 0x59,
    "kp2": 0x5A,
    "kp3": 0x5B,
    "kp4": 0x5C,
    "kp5": 0x5D,
    "kp6": 0x5E,
    "kp7": 0x5F,
    "kp8": 0x60,
    "kp9": 0x61,
    "kp0": 0x62,
    "kpdot": 0x63,
    "102nd": 0x64,
    "compose": 0x65,
    "power": 0x66,
    "kpequal": 0x67,
    "f13": 0x68,
    "f14": 0x69,
    "f15": 0x6A,
    "f16": 0x6B,
    "f17": 0x6C,
    "f18": 0x6D,
    "f19": 0x6E,
    "f20": 0x6F,
    "f21": 0x70,
    "f22": 0x71,
    "f23": 0x72,
    "f24": 0x73,
    "open": 0x74,
    "help": 0x75,
    "props": 0x76,
    "front": 0x77,
    "stop": 0x78,
    "again": 0x79,
    "undo": 0x7A,
    "cut": 0x7B,
    "copy": 0x7C,
    "paste": 0x7D,
    "find": 0x7E,
    "mute": 0x7F,
    "volumeup": 0x80,
    "volumedown": 0x81,
    "kpcomma": 0x85,
    "ro": 0x87,
    "katakanahiragana": 0x88,
    "yen": 0x89,
    "henkan": 0x8A,
    "muhenkan": 0x8B,
    "kpjpcomma": 0x8C,
    "hangeul": 0x90,
    "hanja": 0x91,
    "katakana": 0x92,
    "hiragana": 0x93,
    "zenkakuhankaku": 0x94,
    "kpleftparen": 0xB6,
    "kprightparen": 0xB7,
    "leftctrl": 0xE0,
    "leftshift": 0xE1,
    "leftalt": 0xE2,
    "leftmeta": 0xE3,
    "rightctrl": 0xE4,
    "rightshift": 0xE5,
    "rightalt": 0xE6,
    "rightmeta": 0xE7,
    "mouse_wheelup": 0xE8,
    "mouse_wheeldown": 0xE9,
    "mouse_left": 0xEA,
    "mouse_right": 0xEB,
    "mouse_middle": 0xEC,
    "mouse_fast": 0xED,
    # Vendor Mappings by GPD
    # Gamepad Buttons
    "dpad_up": 0xFF01,
    "dpad_down": 0xFF02,
    "dpad_left": 0xFF03,
    "dpad_right": 0xFF04,
    "btn_a": 0xFF05,
    "btn_b": 0xFF06,
    "btn_x": 0xFF07,
    "btn_y": 0xFF08,
    "ls_up": 0xFF09,
    "ls_down": 0xFF0A,
    "ls_left": 0xFF0B,
    "ls_right": 0xFF0C,
    "ls": 0xFF0D,
    "rs": 0xFF0E,
    "start": 0xFF0F,
    "select": 0xFF10,
    "menu": 0xFF11,
    "lb": 0xFF12,
    "rb": 0xFF13,
    "lt": 0xFF14,
    "rt": 0xFF15,
    "rs_up": 0xFF16,
    "rs_down": 0xFF17,
    "rs_left": 0xFF18,
    "rs_right": 0xFF19,
}


BUTTON_MAP = {
    # Standard buttons (rs is mouse always)
    "dpad_up": 0,
    "dpad_down": 2,
    "dpad_left": 4,
    "dpad_right": 6,
    "a": 8,
    "b": 10,
    "x": 12,
    "y": 14,
    "ls_up": 16,
    "ls_down": 18,
    "ls_left": 20,
    "ls_right": 22,
    "ls": 24,
    "rs": 26,
    "start": 28,
    "select": 30,
    "menu": 32,
    "lb": 34,
    "rb": 36,
    "lt": 38,
    "rt": 40,
    # Macro chains
    "extra_l1": 50,
    "extra_l2": 52,
    "extra_l3": 54,
    "extra_l4": 56,
    "extra_r1": 58,
    "extra_r2": 60,
    "extra_r3": 62,
    "extra_r4": 64,
}

DEADZONE_MAP = {
    "ls_boundary": 72,
    "ls_center": 73,
    "rs_boundary": 74,
    "rs_center": 75,
}

DELAY_MAP = {
    "extra_l1": 80,
    "extra_l2": 82,
    "extra_l3": 84,
    "extra_l4": 86,
    "extra_r1": 88,
    "extra_r2": 90,
    "extra_r3": 92,
    "extra_r4": 94,
}

RGB_MODES = {
    "off": 0,
    "constant": 1,
    "breathed": 0x11,
    "rotated": 0x21,
}

RUMBLE_MODES = {
    "off": 0,
    "medium": 1,
    "high": 2,
}

BACKBUTTONS_HHD = {
    "buttons": {
        "extra_l1": "f20",  # "sysrq",
        "extra_l2": "none",
        "extra_l3": "none",
        "extra_l4": "none",
        "extra_r1": "f21",  # "pause",
        "extra_r2": "none",
        "extra_r3": "none",
        "extra_r4": "none",
    },
    "delays": {
        "extra_l1": 0,
        "extra_l2": 0,
        "extra_l3": 0,
        "extra_l4": 25,
        "extra_r1": 0,
        "extra_r2": 0,
        "extra_r3": 0,
        "extra_r4": 25,
    },
}
BACKBUTTONS_DEFAULT = {
    "buttons": {
        "extra_l1": "sysrq",
        "extra_l2": "none",
        "extra_l3": "none",
        "extra_l4": "none",
        "extra_r1": "pause",
        "extra_r2": "none",
        "extra_r3": "none",
        "extra_r4": "none",
    },
    "delays": {
        "extra_l1": 0,
        "extra_l2": 0,
        "extra_l3": 0,
        "extra_l4": 300,
        "extra_r1": 0,
        "extra_r2": 0,
        "extra_r3": 0,
        "extra_r4": 300,
    },
}

BUTTONS_DEFAULT = {
    "dpad_up": "mouse_wheelup",
    "dpad_down": "mouse_wheeldown",
    "dpad_left": "home",
    "dpad_right": "end",
    "a": "down",
    "b": "right",
    "x": "left",
    "y": "up",
    "ls_up": "w",
    "ls_down": "s",
    "ls_left": "a",
    "ls_right": "d",
    "ls": "space",
    "rs": "enter",
    "start": "none",
    "select": "none",
    "menu": "none",
}

BUTTONS_PHAWX = {
    "dpad_up": "up",
    "dpad_down": "down",
    "dpad_left": "left",
    "dpad_right": "right",
    "a": "space",
    "b": "leftctrl",
    "x": "z",
    "y": "leftalt",
    "ls_up": "w",
    "ls_down": "s",
    "ls_left": "a",
    "ls_right": "d",
    "ls": "leftshift",
    "rs": "enter",
    "start": "escape",
    "select": "enter",
    "menu": "none",
}

BUTTONS_TRIGGERS_DEFAULT = {
    "lb": "mouse_left",
    "rb": "mouse_right",
    "lt": "mouse_middle",
    "rt": "mouse_fast",
}

BUTTONS_TRIGGERS_STEAMOS = {
    "lb": "mouse_middle",
    "rb": "mouse_fast",
    "lt": "mouse_right",
    "rt": "mouse_left",
}

WSIZE = 33
RSIZE = 64


def get_command(cid: int, ofs: int = 0, payload: bytes = b"") -> bytes:
    base = bytes([0x01, 0xA5, cid, 0x5A, 0xFF ^ cid, 0x00, ofs, 0x00]) + payload
    return base + bytes([0x00] * (33 - len(base)))


PAUSE = 0.05

GM_SUPPROTED_VERSIONS = {3: 0x15, 4: 0x09, 5: 0x10}  # Win Max 2  # Win 4  # Win Mini
EXT_SUPPORTED_VERSIONS = {1: 0x23, 4: 0x07, 5: 0x04}


def check_fwver(res: bytes):
    ready = res[8] == 0xAA
    gm_major_ver = res[9]
    gm_minor_ver = res[10]
    ext_major_ver = res[11]
    ext_minor_ver = res[12]

    fwver = f"X{gm_major_ver}{gm_minor_ver:02x}K{ext_major_ver}{ext_minor_ver:02x}"

    # Version check
    for k, v in GM_SUPPROTED_VERSIONS.items():
        if gm_major_ver == k:
            assert (
                gm_minor_ver <= v
            ), f"Unsupported gamepad firmware version {fwver} (up to X{k}{v:02x})"
            break
    else:
        raise ValueError(f"Unsupported gamepad major version {gm_major_ver} in {fwver}")
    for k, v in EXT_SUPPORTED_VERSIONS.items():
        if ext_major_ver == k:
            assert (
                ext_minor_ver <= v
            ), f"Unsupported extendboard firmware {fwver} version (up to K{k}{v:02x})"
            break
    else:
        raise ValueError(
            f"Unsupported extendboard major version {gm_major_ver} in {fwver}"
        )

    return ready, fwver


def read_config(d: Device) -> tuple[str, bytes]:
    ready = False
    while not ready:
        d.send_feature_report(get_command(0x10))
        time.sleep(PAUSE)
        fw_data = d.get_feature_report(0x01)
        time.sleep(PAUSE)
        ready, fwver = check_fwver(fw_data)

    logger.info(f"Device ready, firmware version: {fwver}")

    cfg = bytes()
    for i in range(4):
        d.send_feature_report(get_command(0x11, i))
        time.sleep(PAUSE)
        cfg += d.get_feature_report(0x01)
        time.sleep(PAUSE)

    logger.info("")
    logger.info("Config:")
    tmp = cfg
    i = 0
    while sum(tmp):
        logger.info(f"{i:02d}: {tmp[:16].hex()}")
        tmp = tmp[16:]
        i += 1

    logger.info("")
    d.send_feature_report(get_command(0x12))
    time.sleep(PAUSE)

    fw_data = d.get_feature_report(0x1)
    ready, fwver_new = check_fwver(fw_data)
    assert fwver == fwver_new, "Firmware version changed during read."
    assert ready, "Device not ready after read."

    chash = int.from_bytes(fw_data[24:28], "little")
    logger.info(f"Sum: x{chash:x}")
    # What is this number, it is on 0x10 too
    # chash2 = int.from_bytes(fw_data[28:32], "little")
    # logger.info(f"Ukn: " + str(chash2))
    computed = sum(cfg)
    logger.info(f"Computed Sum: x{computed:x}")

    assert computed == chash, "Did not read config properly: hash mismatch"
    return fwver, cfg


def write_config(d: Device, fwver: str, cfg: bytes):
    d.send_feature_report(get_command(0x21))
    time.sleep(PAUSE)
    fw_data = d.get_feature_report(0x01)
    time.sleep(PAUSE)
    ready = False
    while not ready:
        ready, fwver_new = check_fwver(fw_data)
        assert fwver == fwver_new, "Firmware version changed during write."

    logger.info(f"Firmware version: {fwver}")
    logger.info("")
    logger.info("Writing Config:")
    for i in range(8):
        sls = cfg[16 * i : 16 * i + 16]
        logger.info(f"{i:02d}: {sls.hex()}")
        d.send_feature_report(get_command(0x21, i, sls))

    d.send_feature_report(get_command(0x22))
    time.sleep(PAUSE)
    fw_data = d.get_feature_report(0x1)
    ready, fwver_new = check_fwver(fw_data)
    assert ready, "Device not ready after write."
    assert fwver == fwver_new, "Firmware version changed during write."

    chash = int.from_bytes(fw_data[24:28], "little")
    logger.info("")
    logger.info(f"Sum: x{chash:x}")
    computed = sum(cfg)
    logger.info(f"Computed Sum: x{computed:x}")

    assert computed == chash, "Did not write config properly: hash mismatch"

    logger.info("Writing config to memory and restarting device.")
    d.send_feature_report(get_command(0x23))


def update_config(
    buttons: dict[str, str] = {},
    delays: dict[str, int] = {},
    deadzones: dict[str, int] = {},
    rumble: str | None = None,
    rgb_mode: str | None = None,
    rgb_color: tuple[int, int, int] | None = None,
):
    devs = enumerate_unique(0x2F24, 0x0135, 0xFF00, 0x0001)
    assert devs, "No devices found."

    dev = devs[0]
    with Device(path=dev["path"]) as d:
        fwver, cfg = read_config(d)

    # Apply changes
    init_cfg = cfg
    cfg = bytearray(cfg)

    for k, v in buttons.items():
        assert k in BUTTON_MAP, f"Unknown button {k}"
        assert v in ACTION_MAP, f"Unknown action {v}"
        cfg[BUTTON_MAP[k] : BUTTON_MAP[k] + 2] = ACTION_MAP[v].to_bytes(2, "little")

    for k, v in delays.items():
        assert k in DELAY_MAP, f"Unknown delay {k}"
        cfg[DELAY_MAP[k] : DELAY_MAP[k] + 2] = v.to_bytes(2, "little")

    for k, v in deadzones.items():
        assert k in DEADZONE_MAP, f"Unknown deadzone {k}"
        cfg[DEADZONE_MAP[k] : DEADZONE_MAP[k] + 1] = v.to_bytes(
            1, "little", signed=True
        )

    if rumble is not None:
        assert rumble in RUMBLE_MODES, f"Unknown rumble mode {rumble}"
        cfg[66 : 66 + 2] = RUMBLE_MODES[rumble].to_bytes(2, "little")

    deadzones = {k: min(max(v, -10), 10) for k, v in deadzones.items()}
    for k, v in deadzones.items():
        assert k in DEADZONE_MAP, f"Unknown deadzone {k}"
        cfg[DEADZONE_MAP[k] : DEADZONE_MAP[k] + 1] = v.to_bytes(
            1, "little", signed=True
        )

    if "K4" in fwver:
        # Limit RGB changes to Win 4 with firmware 40X
        if rgb_mode is not None:
            assert rgb_mode in RGB_MODES, f"Unknown rgb mode {rgb_mode}"
            cfg[68] = RGB_MODES[rgb_mode]

        if rgb_color is not None:
            assert len(rgb_color) == 3, "RGB color must be a tuple of 3 integers"
            cfg[69 : 69 + 3] = bytes(rgb_color)

    if all(i == j for i, j in zip(cfg, init_cfg)):
        logger.info("No changes to apply. Skipping write.")
        return fwver

    with Device(path=dev["path"]) as d:
        write_config(d, fwver, bytes(cfg))

    return fwver


def explain_config():
    ACTION_MAP_REV = {v: k for k, v in ACTION_MAP.items()}
    RGB_MODES_REV = {v: k for k, v in RGB_MODES.items()}

    devs = enumerate_unique(0x2F24, 0x0135, 0xFF00, 0x0001)
    assert devs, "No devices found."

    dev = devs[0]
    with Device(path=dev["path"]) as d:
        _, cfg = read_config(d)

    logger.info("\nButtons:")
    for k, v in BUTTON_MAP.items():
        val = int.from_bytes(cfg[v : v + 2], "little")
        val = ACTION_MAP_REV.get(val, f"0x{val:02x}")
        logger.info(f"  {k}: {val}")

    logger.info("\nMacro Delays:")
    for k, v in DELAY_MAP.items():
        val = int.from_bytes(cfg[v : v + 2], "little")
        logger.info(f"  {k}: {val}ms")

    logger.info("\nDeadzones:")
    for k, v in DEADZONE_MAP.items():
        val = int.from_bytes(cfg[v : v + 2], "little", signed=True)
        logger.info(f"  {k}: {val}")

    rumbl = "ukn"
    match int.from_bytes(cfg[66 : 66 + 2], "little"):
        case 0:
            rumbl = "off"
        case 1:
            rumbl = "medium"
        case 2:
            rumbl = "high"
    logger.info(f"\nRumble: {rumbl}")

    rgb_mode = RGB_MODES_REV.get(cfg[68], "ukn")
    rgb_color = cfg[69 : 69 + 3].hex()

    logger.info(f"RGB: {rgb_mode} #{rgb_color}")


# From factory, the following are the default values:
# Buttons:
#   dpad_up: mouse_wheelup
#   dpad_down: mouse_wheeldown
#   dpad_left: home
#   dpad_right: end
#   a: down
#   b: right
#   x: left
#   y: up
#   ls_up: w
#   ls_down: s
#   ls_left: a
#   ls_right: d
#   ls: space
#   rs: enter
#   start: none
#   select: none
#   menu: none
#   lb: mouse_left
#   rb: mouse_right
#   lt: mouse_middle
#   rt: mouse_fast
#   extra_l1: sysrq
#   extra_l2: none
#   extra_l3: none
#   extra_l4: none
#   extra_r1: pause
#   extra_r2: none
#   extra_r3: none
#   extra_r4: none
#
# I might have changed the macro delays
# Macro Delays:
#   extra_l1: 0ms
#   extra_l2: 0ms
#   extra_l3: 0ms
#   extra_l4: 300ms
#   extra_r1: 0ms
#   extra_r2: 0ms
#   extra_r3: 0ms
#   extra_r4: 300ms
#
# Deadzones:
#   ls_deadzone: 0
#   ls_center: 0
#   rs_deadzone: 0
#   rs_center: 0

# Rumble: medium
# RGB: rotated #0000ff

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    update_config(**BACKBUTTONS_HHD)
